"""Tests for `galkin` module."""
import pytest
import numpy.testing as npt
import lenstronomy.Util.multi_gauss_expansion as mge
import numpy as np
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.Analysis.light_profile import LightProfileAnalysis
from lenstronomy.GalKin.galkin import Galkin


class TestGalkin(object):
    def setup_method(self):
        pass

    def test_mge_hernquist_light(self):
        """Compare power-law profiles analytical vs.

        numerical
        :return:
        """
        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        # aperture as slit
        aperture_type = "slit"
        length = 3.8
        width = 0.9
        kwargs_aperture = {
            "length": length,
            "width": width,
            "center_ra": 0,
            "center_dec": 0,
            "angle": 0,
            "aperture_type": aperture_type,
        }

        psf_fwhm = 0.7  # Gaussian FWHM psf
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}

        # light profile
        light_profile_list = ["HERNQUIST"]
        r_eff = 1.8
        kwargs_light = [
            {"Rs": r_eff, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec

        # mge of light profile
        lightModel = LightModel(light_profile_list)
        r_array = np.logspace(-2, 2, 100)
        flux_r = lightModel.surface_brightness(r_array, 0, kwargs_light)
        amps, sigmas, norm = mge.mge_1d(r_array, flux_r, N=20)
        light_profile_list_mge = ["MULTI_GAUSSIAN"]
        kwargs_light_mge = [{"amp": amps, "sigma": sigmas}]
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}

        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_profile = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        kwargs_numerics = {
            "interpol_grid_num": 100,
            "log_integration": True,
            "max_integrate": 100,
            "min_integrate": 0.01,
        }
        galkin = Galkin(
            kwargs_model=kwargs_model,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_aperture=kwargs_aperture,
            kwargs_numerics=kwargs_numerics,
        )
        sigma_v = galkin.dispersion(kwargs_profile, kwargs_light, kwargs_anisotropy)

        kwargs_model_mge = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list_mge,
            "anisotropy_model": anisotropy_type,
        }
        galkin = Galkin(
            kwargs_model=kwargs_model_mge,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_aperture=kwargs_aperture,
            kwargs_numerics=kwargs_numerics,
        )
        sigma_v2 = galkin.dispersion(
            kwargs_profile, kwargs_light_mge, kwargs_anisotropy
        )

        print(sigma_v, sigma_v2, "sigma_v Galkin, sigma_v MGEn")
        print((sigma_v / sigma_v2) ** 2)

        npt.assert_almost_equal((sigma_v - sigma_v2) / sigma_v2, 0, decimal=2)

    def test_mge_power_law_lens(self):
        """Compare power-law profiles analytical vs.

        numerical
        :return:
        """
        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        # aperture as slit
        aperture_type = "slit"
        length = 3.8
        width = 0.9
        kwargs_aperture = {
            "length": length,
            "width": width,
            "center_ra": 0,
            "center_dec": 0,
            "angle": 0,
            "aperture_type": aperture_type,
        }

        psf_fwhm = 0.7  # Gaussian FWHM psf
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}

        # light profile
        light_profile_list = ["HERNQUIST"]
        r_eff = 1.8
        kwargs_light = [
            {"Rs": r_eff, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec

        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_profile = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # mge of lens profile
        lensModel = LensModel(mass_profile_list)
        r_array = np.logspace(-2, 2, 100) * theta_E
        kappa_r = lensModel.kappa(r_array, 0, kwargs_profile)
        amps, sigmas, norm = mge.mge_1d(r_array, kappa_r, N=20)
        mass_profile_list_mge = ["MULTI_GAUSSIAN_KAPPA"]
        kwargs_profile_mge = [{"amp": amps, "sigma": sigmas}]
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        kwargs_numerics = {
            "interpol_grid_num": 100,
            "log_integration": True,
            "max_integrate": 100,
            "min_integrate": 0.01,
        }

        galkin = Galkin(
            kwargs_model=kwargs_model,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_aperture=kwargs_aperture,
            kwargs_numerics=kwargs_numerics,
        )
        sigma_v = galkin.dispersion(kwargs_profile, kwargs_light, kwargs_anisotropy)

        kwargs_model = {
            "mass_profile_list": mass_profile_list_mge,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        galkin = Galkin(
            kwargs_model=kwargs_model,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_aperture=kwargs_aperture,
            kwargs_numerics=kwargs_numerics,
        )
        sigma_v2 = galkin.dispersion(
            kwargs_profile_mge, kwargs_light, kwargs_anisotropy
        )

        print(sigma_v, sigma_v2, "sigma_v Galkin, sigma_v MGEn")
        print((sigma_v / sigma_v2) ** 2)

        npt.assert_almost_equal((sigma_v - sigma_v2) / sigma_v2, 0, decimal=2)

    def test_mge_light_and_mass(self):
        # anisotropy profile
        anisotropy_model = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        # aperture as slit
        aperture_type = "slit"
        length = 3.8
        width = 0.9
        kwargs_aperture = {
            "length": length,
            "width": width,
            "center_ra": 0,
            "center_dec": 0,
            "angle": 0,
            "aperture_type": aperture_type,
        }

        psf_fwhm = 0.7  # Gaussian FWHM psf
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}

        # light profile
        light_profile_list = ["HERNQUIST"]
        r_eff = 1.8
        kwargs_light = [
            {"Rs": r_eff, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec

        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_profile = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # mge of light profile
        lightModel = LightModel(light_profile_list)
        r_array = np.logspace(-2, 2, 200) * r_eff * 2
        flux_r = lightModel.surface_brightness(r_array, 0, kwargs_light)
        amps, sigmas, norm = mge.mge_1d(r_array, flux_r, N=20)
        light_profile_list_mge = ["MULTI_GAUSSIAN"]
        kwargs_light_mge = [{"amp": amps, "sigma": sigmas}]

        # mge of lens profile
        lensModel = LensModel(mass_profile_list)
        r_array = np.logspace(-2, 2, 200)
        kappa_r = lensModel.kappa(r_array, 0, kwargs_profile)
        amps, sigmas, norm = mge.mge_1d(r_array, kappa_r, N=20)
        mass_profile_list_mge = ["MULTI_GAUSSIAN_KAPPA"]
        kwargs_profile_mge = [{"amp": amps, "sigma": sigmas}]
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_model,
        }
        kwargs_numerics = {
            "interpol_grid_num": 100,
            "log_integration": True,
            "max_integrate": 100,
            "min_integrate": 0.01,
        }
        galkin = Galkin(
            kwargs_model=kwargs_model,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_aperture=kwargs_aperture,
            kwargs_numerics=kwargs_numerics,
        )
        sigma_v = galkin.dispersion(kwargs_profile, kwargs_light, kwargs_anisotropy)

        kwargs_model_mge = {
            "mass_profile_list": mass_profile_list_mge,
            "light_profile_list": light_profile_list_mge,
            "anisotropy_model": anisotropy_model,
        }

        galkin = Galkin(
            kwargs_model=kwargs_model_mge,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_aperture=kwargs_aperture,
            kwargs_numerics=kwargs_numerics,
        )
        sigma_v2 = galkin.dispersion(
            kwargs_profile_mge, kwargs_light_mge, kwargs_anisotropy
        )

        print(sigma_v, sigma_v2, "sigma_v Galkin, sigma_v MGEn")
        print((sigma_v / sigma_v2) ** 2)
        npt.assert_almost_equal((sigma_v - sigma_v2) / sigma_v2, 0, decimal=2)

    def test_sersic_vs_hernquist_kinematics(self):
        """
        attention: this test only works for Sersic indices > \approx 2!
        Lower n_sersic will result in different predictions with the Hernquist assumptions
        replacing the correct Light model!
        :return:
        """
        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        # aperture as slit
        aperture_type = "slit"
        length = 3.8
        width = 0.9
        kwargs_aperture = {
            "length": length,
            "width": width,
            "center_ra": 0,
            "center_dec": 0,
            "angle": 0,
            "aperture_type": aperture_type,
        }

        psf_fwhm = 0.7  # Gaussian FWHM psf
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}

        # light profile
        light_profile_list = ["SERSIC"]
        r_sersic = 0.3
        n_sersic = 2.8
        kwargs_light = [
            {
                "amp": 1.0,
                "R_sersic": r_sersic,
                "n_sersic": n_sersic,
                "center_x": 0,
                "center_y": 0,
            }
        ]  # effective half light radius (2d projected) in arcsec

        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_profile = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # Hernquist fit to Sersic profile
        profile_analysis = LightProfileAnalysis(LightModel(["SERSIC"]))
        r_eff = profile_analysis.half_light_radius(
            kwargs_light, grid_spacing=0.1, grid_num=100
        )
        print(r_eff)
        light_profile_list_hernquist = ["HERNQUIST"]
        kwargs_light_hernquist = [{"Rs": r_eff * 0.551, "amp": 1.0}]

        # mge of light profile
        lightModel = LightModel(light_profile_list)
        r_array = np.logspace(-3, 2, 100) * r_eff * 2
        print(r_sersic / r_eff, "r_sersic/r_eff")
        flux_r = lightModel.surface_brightness(r_array, 0, kwargs_light)
        amps, sigmas, norm = mge.mge_1d(r_array, flux_r, N=20)
        light_profile_list_mge = ["MULTI_GAUSSIAN"]
        kwargs_light_mge = [{"amp": amps, "sigma": sigmas}]
        print(amps, sigmas, "amp", "sigma")
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}

        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list_hernquist,
            "anisotropy_model": anisotropy_type,
        }

        galkin = Galkin(
            kwargs_model=kwargs_model,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_aperture=kwargs_aperture,
            kwargs_numerics={},
        )
        sigma_v = galkin.dispersion(
            kwargs_profile, kwargs_light_hernquist, kwargs_anisotropy
        )

        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list_mge,
            "anisotropy_model": anisotropy_type,
        }
        galkin = Galkin(
            kwargs_model=kwargs_model,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_aperture=kwargs_aperture,
            kwargs_numerics={},
        )
        sigma_v2 = galkin.dispersion(
            kwargs_profile, kwargs_light_mge, kwargs_anisotropy
        )

        print(sigma_v, sigma_v2, "sigma_v Galkin, sigma_v MGEn")
        print((sigma_v / sigma_v2) ** 2)

        npt.assert_almost_equal((sigma_v - sigma_v2) / sigma_v2, 0, decimal=1)


if __name__ == "__main__":
    pytest.main()



from lenstronomy.LensModel.Util.epl_util import brentq_nojit
import numpy as np
import numpy.testing as npt


def test_brentq_nojit():
    npt.assert_almost_equal(
        brentq_nojit(lambda x, args: np.sin(x), np.pi / 2, 3 * np.pi / 2),
        np.pi,
        decimal=10,
    )
    npt.assert_almost_equal(
        brentq_nojit(lambda x, args: np.cos(x), np.pi, np.pi / 2), np.pi / 2, decimal=10
    )



import numpy as np

__all__ = ["NFWParam"]


class NFWParam(object):
    """Class which contains a halo model parameters dependent on cosmology for NFW
    profile All distances are given in physical units.

    Mass definitions are relative to 200 crit including redshift evolution. The redshift
    evolution is cosmology dependent (dark energy). The H0 dependence is propagated into
    the input and return units.
    """

    rhoc = 2.77536627e11  # critical density [h^2 M_sun Mpc^-3]

    def __init__(self, cosmo=None):
        """

        :param cosmo: astropy.cosmology instance
        """
        from astropy.cosmology import default_cosmology

        if cosmo is None:
            cosmo = default_cosmology.get()
        self.cosmo = cosmo

    def rhoc_z(self, z):
        """

        :param z: redshift
        :return: critical density of the universe at redshift z in physical units [h^2 M_sun Mpc^-3]
        """
        return self.rhoc * (self.cosmo.efunc(z)) ** 2
        # return self.rhoc*(1+z)**3

    @staticmethod
    def M200(rs, rho0, c):
        """Calculation of the mass enclosed r_200 for NFW profile defined as.

        .. math::
            M_{200} = 4 \\pi \\rho_0^{3} * \\left(\\log(1+c) - c / (1 + c)  \\right))

        :param rs: scale radius
        :type rs: float
        :param rho0: density normalization (characteristic density) in units mass/[distance unit of rs]^3
        :type rho0: float
        :param c: concentration
        :type c: float [4,40]
        :return: M(R_200) mass in units of rho0 * rs^3
        """
        return 4 * np.pi * rho0 * rs**3 * (np.log(1.0 + c) - c / (1.0 + c))

    def r200_M(self, M, z):
        """Computes the radius R_200 crit of a halo of mass M in physical mass M/h.

        :param M: halo mass in M_sun/h
        :type M: float or numpy array
        :param z: redshift
        :type z: float
        :return: radius R_200 in physical Mpc/h
        """
        return (3 * M / (4 * np.pi * self.rhoc_z(z) * 200)) ** (1.0 / 3.0)

    def M_r200(self, r200, z):
        """

        :param r200: r200 in physical Mpc/h
        :param z: redshift
        :return: M200 in M_sun/h
        """
        return self.rhoc_z(z) * 200 * r200**3 * 4 * np.pi / 3.0

    def rho0_c(self, c, z):
        """Computes density normalization as a function of concentration parameter.

        :param c: concentration
        :param z: redshift
        :return: density normalization in h^2/Mpc^3 (physical)
        """
        return 200.0 / 3 * self.rhoc_z(z) * c**3 / (np.log(1.0 + c) - c / (1.0 + c))

    def c_rho0(self, rho0, z):
        """Computes the concentration given density normalization rho_0 in h^2/Mpc^3
        (physical) (inverse of function rho0_c)

        :param rho0: density normalization in h^2/Mpc^3 (physical)
        :param z: redshift
        :return: concentration parameter c
        """
        if not hasattr(self, "_c_rho0_interp"):
            c_array = np.linspace(0.1, 30, 100)
            rho0_array = self.rho0_c(c_array, z)
            from scipy import interpolate

            self._c_rho0_interp = interpolate.InterpolatedUnivariateSpline(
                rho0_array, c_array, w=None, bbox=[None, None], k=3
            )
        return self._c_rho0_interp(rho0)

    @staticmethod
    def c_M_z(M, z):
        """
        fitting function of http://moriond.in2p3.fr/J08/proceedings/duffy.pdf for the mass and redshift dependence of
        the concentration parameter

        :param M: halo mass in M_sun/h
        :type M: float or numpy array
        :param z: redshift
        :type z: float >0
        :return: concentration parameter as float
        """
        # fitted parameter values
        A = 5.22
        B = -0.072
        C = -0.42
        M_pivot = 2.0 * 10**12
        return A * (M / M_pivot) ** B * (1 + z) ** C

    def nfw_Mz(self, M, z):
        """Returns all needed parameter (in physical units modulo h) to draw the profile
        of the main halo r200 in physical Mpc/h rho_s in  h^2/Mpc^3 (physical) Rs in
        Mpc/h physical c unit less.

        :param M: Mass in physical M_sun/h
        :param z: redshift
        """
        c = self.c_M_z(M, z)
        r200 = self.r200_M(M, z)
        rho0 = self.rho0_c(c, z)
        Rs = r200 / c
        return r200, rho0, c, Rs



__author__ = "sibirrer"

__all__ = ["SpecialParam"]

import numpy as np
from .param_group import ModelParamGroup, SingleParam, ArrayParam


# ==================================== #
# == Defining individual parameters == #
# ==================================== #


class DdtSamplingParam(SingleParam):
    """Time delay parameter."""

    param_names = ["D_dt"]
    _kwargs_lower = {"D_dt": 0}
    _kwargs_upper = {"D_dt": 100000}


class DdSamplingParam(SingleParam):
    """Deflector distance parameter."""

    param_names = ["D_d"]
    _kwargs_lower = {"D_d": 0}
    _kwargs_upper = {"D_d": 100000}


class BetaAnisotropyParam(SingleParam):
    """Cylindrical anisotropy parameter."""

    param_names = ["b_ani"]
    _kwargs_lower = {"b_ani": -1}
    _kwargs_upper = {"b_ani": 1}


class InclinationParam(SingleParam):
    """Inclination parameter (radians)"""

    param_names = ["incli"]
    _kwargs_lower = {"incli": 0}
    _kwargs_upper = {"incli": np.pi / 2}


class SourceSizeParam(SingleParam):
    """Source size parameter."""

    param_names = ["source_size"]
    _kwargs_lower = {"source_size": 0}
    _kwargs_upper = {"source_size": 1}


class SourceGridOffsetParam(SingleParam):
    """Source grid offset, both x and y."""

    param_names = ["delta_x_source_grid", "delta_y_source_grid"]
    _kwargs_lower = {"delta_x_source_grid": -100, "delta_y_source_grid": -100}
    _kwargs_upper = {"delta_x_source_grid": 100, "delta_y_source_grid": 100}


class MassScalingParam(ArrayParam):
    """Mass scaling.

    Can scale the masses of arbitrary subsets of lens models
    """

    _kwargs_lower = {"scale_factor": 0}
    _kwargs_upper = {"scale_factor": 1000}

    def __init__(self, num_scale_factor):
        super().__init__(on=int(num_scale_factor) > 0)
        self.param_names = {"scale_factor": int(num_scale_factor)}


class PointSourceOffsetParam(ArrayParam):
    """Point source offset, both x and y."""

    _kwargs_lower = {"delta_x_image": -1, "delta_y_image": -1}
    _kwargs_upper = {"delta_x_image": 1, "delta_y_image": 1}

    def __init__(self, offset, num_images):
        super().__init__(on=offset and (int(num_images) > 0))
        self.param_names = {
            "delta_x_image": int(num_images),
            "delta_y_image": int(num_images),
        }


class Tau0ListParam(ArrayParam):
    """Optical depth renormalization parameters."""

    _kwargs_lower = {"tau0_list": 0}
    _kwargs_upper = {"tau0_list": 1000}

    def __init__(self, num_tau0):
        super().__init__(on=int(num_tau0) > 0)
        self.param_names = {"tau0_list": int(num_tau0)}


class ZSamplingParam(ArrayParam):
    """Redshift sampling."""

    _kwargs_lower = {"z_sampling": 0}
    _kwargs_upper = {"z_sampling": 1000}

    def __init__(self, num_z_sampling):
        super().__init__(on=int(num_z_sampling) > 0)
        self.param_names = {"z_sampling": int(num_z_sampling)}


class GeneralScalingParam(ArrayParam):
    """General lens scaling.

    For each scaled lens parameter, adds a `{param}_scale_factor` and
    `{param}_scale_pow` special parameter, and updates the scaled param as `param =
    param_scale_factor * param**param_scale_pow`.
    """

    def __init__(self, params: dict):
        # params is a dictionary
        self.param_names = {}
        self._kwargs_lower = {}
        self._kwargs_upper = {}

        super().__init__(params)
        if not self.on:
            return

        for name, array in params.items():
            num_param = np.max(array)

            if num_param > 0:
                fac_name = f"{name}_scale_factor"
                self.param_names[fac_name] = num_param
                self._kwargs_lower[fac_name] = 0
                self._kwargs_upper[fac_name] = 1000

                pow_name = f"{name}_scale_pow"
                self.param_names[pow_name] = num_param
                self._kwargs_lower[pow_name] = -10
                self._kwargs_upper[pow_name] = 10


# ======================================== #
# == All together: Composing into class == #
# ======================================== #


class SpecialParam(object):
    """Class that handles special parameters that are not directly part of a specific
    model component.

    These includes cosmology relevant parameters, astrometric errors and overall scaling
    parameters.
    """

    def __init__(
        self,
        Ddt_sampling=False,
        mass_scaling=False,
        num_scale_factor=1,
        general_scaling_params=None,
        kwargs_fixed=None,
        kwargs_lower=None,
        kwargs_upper=None,
        point_source_offset=False,
        source_size=False,
        num_images=0,
        num_tau0=0,
        num_z_sampling=0,
        source_grid_offset=False,
        kinematic_sampling=False,
    ):
        """

        :param Ddt_sampling: bool, if True, samples the time-delay distance D_dt (in units of Mpc)
        :param mass_scaling: bool, if True, samples a mass scaling factor between different profiles
        :param num_scale_factor: int, number of independent mass scaling factors being sampled
        :param kwargs_fixed: keyword arguments, fixed parameters during sampling
        :param kwargs_lower: keyword arguments, lower bound of parameters being sampled
        :param kwargs_upper: keyword arguments, upper bound of parameters being sampled
        :param point_source_offset: bool, if True, adds relative offsets ot the modeled image positions relative to the
         time-delay and lens equation solver
        :param num_images: number of point source images such that the point source offset parameters match their
         numbers
        :param source_size: bool, if True, samples a source size parameters to be evaluated in the flux ratio likelihood
        :param num_tau0: integer, number of different optical depth re-normalization factors
        :param num_z_sampling: integer, number of different lens redshifts to be sampled
        :param source_grid_offset: bool, if True, samples two parameters (x, y) for the offset of the pixelated source
         plane grid coordinates.
         Warning: this is only defined for pixel-based source modelling (e.g. 'SLIT_STARLETS' light profile)
        :param kinematic_sampling: bool, if True, samples the kinematic parameters b_ani, incli, with cosmography
         D_dt (overrides _D_dt_sampling) and Dd
        """

        self._D_dt_sampling = DdtSamplingParam(Ddt_sampling or kinematic_sampling)

        self._D_d_sampling = DdSamplingParam(kinematic_sampling)
        self._b_ani_sampling = BetaAnisotropyParam(kinematic_sampling)
        self._incli_sampling = InclinationParam(kinematic_sampling)
        if not mass_scaling:
            num_scale_factor = 0
        self._mass_scaling = MassScalingParam(num_scale_factor)

        self._general_scaling = GeneralScalingParam(general_scaling_params or dict())

        if point_source_offset:
            self._point_source_offset = PointSourceOffsetParam(True, num_images)
        else:
            self._point_source_offset = PointSourceOffsetParam(False, 0)
        self._source_size = SourceSizeParam(source_size)
        self._tau0 = Tau0ListParam(num_tau0)
        self._z_sampling = ZSamplingParam(num_z_sampling)
        self._source_grid_offset = SourceGridOffsetParam(source_grid_offset)

        if kwargs_fixed is None:
            kwargs_fixed = {}
        self._kwargs_fixed = kwargs_fixed

        if kwargs_lower is None:
            kwargs_lower = {}
            for group in self._param_groups:
                kwargs_lower = dict(kwargs_lower, **group.kwargs_lower)
        if kwargs_upper is None:
            kwargs_upper = {}
            for group in self._param_groups:
                kwargs_upper = dict(kwargs_upper, **group.kwargs_upper)

        self.lower_limit = kwargs_lower
        self.upper_limit = kwargs_upper

    def get_params(self, args, i, impose_bound=False):
        """

        :param args: argument list
        :param i: integer, list index to start the read out for this class
        :param impose_bound: bool, if True, imposes the lower and upper limits on the sampled parameters
        :return: keyword arguments related to args, index after reading out arguments of this class
        """
        if impose_bound:
            result = ModelParamGroup.compose_get_params(
                self._param_groups,
                args,
                i,
                kwargs_fixed=self._kwargs_fixed,
                kwargs_lower=self.lower_limit,
                kwargs_upper=self.upper_limit,
            )
        else:
            result = ModelParamGroup.compose_get_params(
                self._param_groups, args, i, kwargs_fixed=self._kwargs_fixed
            )
        return result

    def set_params(self, kwargs_special):
        """

        :param kwargs_special: keyword arguments with parameter settings
        :return: argument list of the sampled parameters extracted from kwargs_special
        """
        return ModelParamGroup.compose_set_params(
            self._param_groups, kwargs_special, kwargs_fixed=self._kwargs_fixed
        )

    def num_param(self):
        """

        :return: integer, number of free parameters sampled (and managed) by this class, parameter names (list of strings)
        """
        return ModelParamGroup.compose_num_params(
            self._param_groups, kwargs_fixed=self._kwargs_fixed
        )

    @property
    def _param_groups(self):
        return [
            self._D_dt_sampling,
            self._D_d_sampling,
            self._b_ani_sampling,
            self._incli_sampling,
            self._mass_scaling,
            self._general_scaling,
            self._point_source_offset,
            self._source_size,
            self._tau0,
            self._z_sampling,
            self._source_grid_offset,
        ]



import numpy as np
from lenstronomy.Data.coord_transforms import Coordinates
from lenstronomy.Data.angular_sensitivity import AngularSensitivity

__all__ = ["PixelGrid"]


class PixelGrid(Coordinates, AngularSensitivity):
    """Class that manages a specified pixel grid (rectangular at the moment) and its
    coordinates."""

    def __init__(
        self,
        nx,
        ny,
        transform_pix2angle,
        ra_at_xy_0,
        dec_at_xy_0,
        antenna_primary_beam=None,
    ):
        """

        :param nx: number of pixels in x-axis
        :param ny: number of pixels in y-axis
        :param transform_pix2angle: 2x2 matrix, mapping of pixel to coordinate
        :param ra_at_xy_0: ra coordinate at pixel (0,0)
        :param dec_at_xy_0: dec coordinate at pixel (0,0)
        :param antenna_primary_beam: 2d numpy array with the same size of imaga_data;
         more descriptions of the primary beam can be found in the AngularSensitivity class
        """
        super(PixelGrid, self).__init__(transform_pix2angle, ra_at_xy_0, dec_at_xy_0)
        self._nx = nx
        self._ny = ny
        self._x_grid, self._y_grid = self.coordinate_grid(nx, ny)
        if antenna_primary_beam is not None:
            pbx, pby = np.shape(antenna_primary_beam)
            if (pbx, pby) != (nx, ny):
                raise ValueError(
                    "The primary beam should have the same size with the image data!"
                )
        AngularSensitivity.__init__(self, antenna_primary_beam)

    @property
    def num_pixel(self):
        """

        :return: number of pixels in the data
        """
        return self._nx * self._ny

    @property
    def num_pixel_axes(self):
        """

        :return: number of pixels per axis, nx ny
        """
        return self._nx, self._ny

    @property
    def width(self):
        """

        :return: width of data frame
        """
        return self._nx * self.pixel_width, self._ny * self.pixel_width

    @property
    def center(self):
        """

        :return: center_x, center_y of coordinate system
        """
        return np.mean(self._x_grid), np.mean(self._y_grid)

    def shift_coordinate_system(self, x_shift, y_shift, pixel_unit=False):
        """Shifts the coordinate system :param x_shift: shift in x (or RA) :param
        y_shift: shift in y (or DEC) :param pixel_unit: bool, if True, units of pixels
        in input, otherwise RA/DEC :return: updated data class with change in coordinate
        system."""
        self._shift_coordinates(x_shift, y_shift, pixel_unit=pixel_unit)
        self._x_grid, self._y_grid = self.coordinate_grid(self._nx, self._ny)

    @property
    def pixel_coordinates(self):
        """

        :return: RA coords, DEC coords
        """
        return self._x_grid, self._y_grid



import unittest
from lenstronomy.SimulationAPI.ObservationConfig.Euclid import Euclid
from lenstronomy.SimulationAPI.observation_api import Instrument, SingleBand
import lenstronomy.Util.util as util


class TestEuclid(unittest.TestCase):
    def setUp(self):
        self.VIS = Euclid()

        kwargs_VIS = self.VIS.kwargs_single_band()

        self.VIS_band = SingleBand(**kwargs_VIS)

        # dictionaries mapping Euclid kwargs to SingleBand kwargs
        self.camera_settings = {
            "read_noise": "_read_noise",
            "pixel_scale": "pixel_scale",
            "ccd_gain": "ccd_gain",
        }
        self.obs_settings = {
            "exposure_time": "_exposure_time",
            "sky_brightness": "_sky_brightness_",
            "magnitude_zero_point": "_magnitude_zero_point",
            "num_exposures": "_num_exposures",
            "seeing": "_seeing",
            "psf_type": "_psf_type",
        }

        self.instrument = Instrument(**self.VIS.camera)

    def test_Euclid_class(self):
        default = self.VIS
        explicit = Euclid(band="VIS")
        self.assertEqual(explicit.camera, default.camera)
        self.assertEqual(explicit.obs, default.obs)

        with self.assertRaises(ValueError):
            bad_band = Euclid(band="g")

        with self.assertRaises(ValueError):
            bad_psf = Euclid(psf_type="pixel")

        single_year = Euclid(coadd_years=2)
        self.assertEqual(single_year.obs["num_exposures"], 1)
        with self.assertRaises(ValueError):
            bad_coadd_years = Euclid(coadd_years=7)

    def test_Euclid_camera(self):
        # comparing camera settings in Euclid instance with those in Instrument instance
        for config, setting in self.camera_settings.items():
            self.assertEqual(
                self.VIS.camera[config],
                getattr(self.instrument, setting),
                msg=f"{config} did not match",
            )

    def test_Euclid_obs(self):
        # comparing obs settings in Euclid instance with those in SingleBand instance
        for config, setting in self.obs_settings.items():
            self.assertEqual(
                self.VIS.obs[config],
                getattr(self.VIS_band, setting),
                msg=f"{config} did not match",
            )

    def test_kwargs_single_band(self):
        kwargs_VIS = util.merge_dicts(self.VIS.camera, self.VIS.obs)
        self.assertEqual(self.VIS.kwargs_single_band(), kwargs_VIS)


if __name__ == "__main__":
    unittest.main()



import unittest
from lenstronomy.SimulationAPI.ObservationConfig.JWST import JWST
from lenstronomy.SimulationAPI.observation_api import Instrument, SingleBand
import lenstronomy.Util.util as util


class TestJWST(unittest.TestCase):
    def setUp(self):
        self.F200W = JWST()  # default is F200W
        self.F356W = JWST(band="F356W")
        self.F356W2 = JWST(band="F356W", psf_type="GAUSSIAN")

        kwargs_F200W = self.F200W.kwargs_single_band()
        kwargs_F356W = self.F356W.kwargs_single_band()
        kwargs_F356W2 = self.F356W2.kwargs_single_band()

        self.F200W_band = SingleBand(**kwargs_F200W)
        self.F356W_band = SingleBand(**kwargs_F356W)
        self.F356W2_band = SingleBand(**kwargs_F356W2)

        # dictionaries mapping JWST kwargs to SingleBand kwargs
        self.camera_settings = {
            "read_noise": "_read_noise",
            "pixel_scale": "pixel_scale",
            "ccd_gain": "ccd_gain",
        }
        self.obs_settings = {
            "exposure_time": "_exposure_time",
            "sky_brightness": "_sky_brightness_",
            "magnitude_zero_point": "_magnitude_zero_point",
            "num_exposures": "_num_exposures",
            "seeing": "_seeing",
            "psf_type": "_psf_type",
        }

        self.instrument = Instrument(**self.F200W.camera)

    def test_JWST_class(self):
        default = self.F200W
        explicit_F200W = JWST(band="F200W", psf_type="PIXEL")
        self.assertEqual(explicit_F200W.camera, default.camera)
        self.assertEqual(explicit_F200W.obs, default.obs)

        with self.assertRaises(ValueError):
            bad_band = JWST(band="g")

        with self.assertRaises(ValueError):
            bad_psf = JWST(psf_type="blah")

        with self.assertRaises(ValueError):
            bad_coadd_years = JWST(coadd_years=100)

    def test_JWST_camera(self):
        # comparing camera settings in JWST instance with those in Instrument instance
        for config, setting in self.camera_settings.items():
            self.assertEqual(
                self.F200W.camera[config],
                getattr(self.instrument, setting),
                msg=f"{config} did not match",
            )

    def test_JWST_obs(self):
        # comparing obs settings in JWST instance with those in SingleBand instance
        for config, setting in self.obs_settings.items():
            self.assertEqual(
                self.F200W.obs[config],
                getattr(self.F200W_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.F356W.obs[config],
                getattr(self.F356W_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.F356W2.obs[config],
                getattr(self.F356W2_band, setting),
                msg=f"{config} did not match",
            )

    def test_kwargs_single_band(self):
        kwargs_F200W = util.merge_dicts(self.F200W.camera, self.F200W.obs)
        self.assertEqual(self.F200W.kwargs_single_band(), kwargs_F200W)


if __name__ == "__main__":
    unittest.main()



from lenstronomy.PointSource.Types.lensed_position import LensedPositions
from lenstronomy.LensModel.lens_model import LensModel
import pytest
import numpy.testing as npt


class TestLensedPosition(object):
    def setup_method(self):
        lens_model = LensModel(lens_model_list=["SIS"])
        self.kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        self.ps_mag = LensedPositions(
            lens_model=lens_model,
            fixed_magnification=True,
            point_source_frame_list=[0, 0],
            index_lens_model_list=[[0]],
        )
        self.ps = LensedPositions(lens_model=lens_model, fixed_magnification=False)
        self.ps_add = LensedPositions(
            lens_model=lens_model, fixed_magnification=[False], additional_images=True
        )
        self.kwargs = {"point_amp": [2, 1], "ra_image": [0, 1.2], "dec_image": [0, 0]}
        self.kwargs_mag = {"source_amp": 2, "ra_image": [0, 1.2], "dec_image": [0, 0]}

    def test_image_source_position(self):
        x_img, y_img = self.ps.image_position(self.kwargs, self.kwargs_lens)
        npt.assert_almost_equal(x_img, self.kwargs["ra_image"])

        x_img_add, y_img_add = self.ps_add.image_position(self.kwargs, self.kwargs_lens)
        print(x_img_add, x_img)
        assert x_img[0] != x_img_add[0]

        # check whether the source solution matches
        x_src, y_src = self.ps.source_position(self.kwargs, self.kwargs_lens)
        lens_model = LensModel(lens_model_list=["SIS"])
        x_src_true, y_src_true = lens_model.ray_shooting(
            x_img_add, y_img_add, kwargs=self.kwargs_lens
        )
        npt.assert_almost_equal(x_src_true[0], x_src_true[1])
        npt.assert_almost_equal(y_src_true[0], y_src_true[1])

        npt.assert_almost_equal(x_src_true, x_src)
        npt.assert_almost_equal(y_src_true, y_src)

    def test_image_amplitude(self):
        amp = self.ps.image_amplitude(
            self.kwargs,
            kwargs_lens=self.kwargs_lens,
            x_pos=self.kwargs["ra_image"],
            y_pos=self.kwargs["dec_image"],
            magnification_limit=None,
            kwargs_lens_eqn_solver=None,
        )
        npt.assert_almost_equal(self.kwargs["point_amp"], amp)

        amp = self.ps_mag.image_amplitude(
            self.kwargs_mag,
            kwargs_lens=self.kwargs_lens,
            x_pos=None,
            y_pos=None,
            magnification_limit=None,
            kwargs_lens_eqn_solver=None,
        )

        amp_pos = self.ps_mag.image_amplitude(
            self.kwargs_mag,
            kwargs_lens=self.kwargs_lens,
            x_pos=self.kwargs["ra_image"],
            y_pos=self.kwargs["dec_image"],
            magnification_limit=None,
            kwargs_lens_eqn_solver=None,
        )
        npt.assert_almost_equal(amp, amp_pos)

    def test_source_amplitude(self):
        amp = self.ps.source_amplitude(self.kwargs, kwargs_lens=self.kwargs_lens)
        amp_mag = self.ps_mag.source_amplitude(
            self.kwargs_mag, kwargs_lens=self.kwargs_lens
        )
        npt.assert_almost_equal(amp_mag, self.kwargs_mag["source_amp"])
        assert amp != amp_mag


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import pytest
import unittest
import lenstronomy.Util.util as util
import lenstronomy.Util.param_util as param_util
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.LightModel.Profiles.gaussian import Gaussian


class TestLightModel(object):
    """Tests the source model routines."""

    def setup_method(self):
        self.light_model_list = [
            "GAUSSIAN",
            "MULTI_GAUSSIAN",
            "SERSIC",
            "SERSIC_ELLIPSE",
            "CORE_SERSIC",
            "SHAPELETS",
            "HERNQUIST",
            "HERNQUIST_ELLIPSE",
            "PJAFFE",
            "PJAFFE_ELLIPSE",
            "UNIFORM",
            "POWER_LAW",
            "NIE",
            "INTERPOL",
            "SHAPELETS_POLAR_EXP",
            "ELLIPSOID",
        ]
        phi_G, q = 0.5, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        self.kwargs = [
            {"amp": 1.0, "sigma": 1.0, "center_x": 0, "center_y": 0},  # 'GAUSSIAN'
            {
                "amp": [1.0, 2],
                "sigma": [1, 3],
                "center_x": 0,
                "center_y": 0,
            },  # 'MULTI_GAUSSIAN'
            {
                "amp": 1,
                "R_sersic": 0.5,
                "n_sersic": 1,
                "center_x": 0,
                "center_y": 0,
            },  # 'SERSIC'
            {
                "amp": 1,
                "R_sersic": 0.5,
                "n_sersic": 1,
                "e1": e1,
                "e2": e2,
                "center_x": 0,
                "center_y": 0,
            },  # 'SERSIC_ELLIPSE'
            {
                "amp": 1,
                "R_sersic": 0.5,
                "Rb": 0.1,
                "gamma": 2.0,
                "n_sersic": 1,
                "e1": e1,
                "e2": e2,
                "center_x": 0,
                "center_y": 0,
            },
            # 'CORE_SERSIC'
            {
                "amp": [1, 1, 1],
                "beta": 0.5,
                "n_max": 1,
                "center_x": 0,
                "center_y": 0,
            },  # 'SHAPELETS'
            {"amp": 1, "Rs": 0.5, "center_x": 0, "center_y": 0},  # 'HERNQUIST'
            {
                "amp": 1,
                "Rs": 0.5,
                "center_x": 0,
                "center_y": 0,
                "e1": e1,
                "e2": e2,
            },  # 'HERNQUIST_ELLIPSE'
            {"amp": 1, "Ra": 1, "Rs": 0.5, "center_x": 0, "center_y": 0},  # 'PJAFFE'
            {
                "amp": 1,
                "Ra": 1,
                "Rs": 0.5,
                "center_x": 0,
                "center_y": 0,
                "e1": e1,
                "e2": e2,
            },  # 'PJAFFE_ELLIPSE'
            {"amp": 1},  # 'UNIFORM'
            {
                "amp": 1.0,
                "gamma": 2.0,
                "e1": e1,
                "e2": e2,
                "center_x": 0,
                "center_y": 0,
            },  # 'POWER_LAW'
            {
                "amp": 0.001,
                "e1": 0,
                "e2": 1.0,
                "center_x": 0,
                "center_y": 0,
                "s_scale": 1.0,
            },  # 'NIE'
            {
                "image": np.zeros((20, 5)),
                "scale": 1,
                "phi_G": 0,
                "center_x": 0,
                "center_y": 0,
            },
            {"amp": [1], "n_max": 0, "beta": 1, "center_x": 0, "center_y": 0},
            {
                "amp": 1,
                "radius": 1.0,
                "e1": 0,
                "e2": 0.1,
                "center_x": 0,
                "center_y": 0,
            },  # 'ELLIPSOID'
        ]

        self.LightModel = LightModel(
            light_model_list=self.light_model_list, sersic_major_axis=False
        )

    def test_init(self):
        model_list = [
            "CORE_SERSIC",
            "SHAPELETS",
            "SHAPELETS_ELLIPSE",
            "SHAPELETS_POLAR",
            "SHAPELETS_POLAR_EXP",
            "UNIFORM",
            "CHAMELEON",
            "DOUBLE_CHAMELEON",
            "TRIPLE_CHAMELEON",
        ]
        lightModel = LightModel(light_model_list=model_list)
        assert len(lightModel.profile_type_list) == len(model_list)

    def test_surface_brightness(self):
        output = self.LightModel.surface_brightness(
            x=1.0, y=1.0, kwargs_list=self.kwargs
        )
        npt.assert_almost_equal(output, 2.647127, decimal=6)

    def test_surface_brightness_array(self):
        output = self.LightModel.surface_brightness(
            x=[1], y=[1], kwargs_list=self.kwargs
        )
        npt.assert_almost_equal(output[0], 2.647127113888489, decimal=6)

    def test_functions_split(self):
        output = self.LightModel.functions_split(x=1.0, y=1.0, kwargs_list=self.kwargs)
        npt.assert_almost_equal(output[0][0], 0.058549831524319168, decimal=6)

    def test_param_name_list(self):
        param_name_list = self.LightModel.param_name_list
        assert len(self.light_model_list) == len(param_name_list)

    def test_param_name_list_latex(self):
        param_name_list = self.LightModel.param_name_list_latex
        assert len(self.light_model_list) == len(param_name_list)

    def test_num_param_linear(self):
        num = self.LightModel.num_param_linear(self.kwargs, list_return=False)
        assert num == 19

        num_list = self.LightModel.num_param_linear(self.kwargs, list_return=True)
        assert num_list[0] == 1

    def test_update_linear(self):
        response, n = self.LightModel.functions_split(1, 1, self.kwargs)
        param = np.ones(n) * 2
        kwargs_out, i = self.LightModel.update_linear(
            param, i=0, kwargs_list=self.kwargs
        )
        assert i == n
        assert kwargs_out[0]["amp"] == 2

    def test_total_flux(self):
        light_model_list = [
            "SERSIC",
            "SERSIC_ELLIPSE",
            "INTERPOL",
            "GAUSSIAN",
            "GAUSSIAN_ELLIPSE",
            "MULTI_GAUSSIAN",
            "MULTI_GAUSSIAN_ELLIPSE",
        ]
        kwargs_list = [
            {
                "amp": 1,
                "R_sersic": 0.5,
                "n_sersic": 1,
                "center_x": 0,
                "center_y": 0,
            },  # 'SERSIC'
            {
                "amp": 1,
                "R_sersic": 0.5,
                "n_sersic": 1,
                "e1": 0.1,
                "e2": 0,
                "center_x": 0,
                "center_y": 0,
            },  # 'SERSIC_ELLIPSE'
            {
                "image": np.ones((20, 5)),
                "scale": 1,
                "phi_G": 0,
                "center_x": 0,
                "center_y": 0,
            },  # 'INTERPOL'
            {"amp": 2, "sigma": 2, "center_x": 0, "center_y": 0},  # 'GAUSSIAN'
            {
                "amp": 2,
                "sigma": 2,
                "e1": 0.1,
                "e2": 0,
                "center_x": 0,
                "center_y": 0,
            },  # 'GAUSSIAN_ELLIPSE'
            {
                "amp": [1, 1],
                "sigma": [2, 1],
                "center_x": 0,
                "center_y": 0,
            },  # 'MULTI_GAUSSIAN'
            {
                "amp": [1, 1],
                "sigma": [2, 1],
                "e1": 0.1,
                "e2": 0,
                "center_x": 0,
                "center_y": 0,
            },  # 'MULTI_GAUSSIAN_ELLIPSE'
        ]
        lightModel = LightModel(light_model_list=light_model_list)
        total_flux_list = lightModel.total_flux(kwargs_list)
        assert total_flux_list[2] == 100
        assert total_flux_list[3] == 2
        assert total_flux_list[4] == 2
        assert total_flux_list[5] == 2
        assert total_flux_list[6] == 2

        total_flux_list = lightModel.total_flux(kwargs_list, norm=True)
        assert total_flux_list[2] == 100
        assert total_flux_list[3] == 1
        assert total_flux_list[4] == 1
        assert total_flux_list[5] == 2
        assert total_flux_list[6] == 2

    def test_delete_interpol_caches(self):
        x, y = util.make_grid(numPix=20, deltapix=1.0)
        gauss = Gaussian()
        flux = gauss.function(x, y, amp=1.0, center_x=0.0, center_y=0.0, sigma=1.0)
        image = util.array2image(flux)

        light_model_list = ["INTERPOL", "INTERPOL"]
        kwargs_list = [
            {"image": image, "scale": 1, "phi_G": 0, "center_x": 0, "center_y": 0},
            {"image": image, "scale": 1, "phi_G": 0, "center_x": 0, "center_y": 0},
        ]
        lightModel = LightModel(light_model_list=light_model_list)
        output = lightModel.surface_brightness(x, y, kwargs_list)
        for func in lightModel.func_list:
            assert hasattr(func, "_image_interp")
        lightModel.delete_interpol_caches()
        for func in lightModel.func_list:
            assert not hasattr(func, "_image_interp")

    def test_check_positive_flux_profile(self):
        ligthModel = LightModel(light_model_list=["GAUSSIAN"])
        kwargs_list = [{"amp": 0, "sigma": 1}]
        bool = ligthModel.check_positive_flux_profile(kwargs_list)
        assert bool

        kwargs_list = [{"amp": -1, "sigma": 1}]
        bool = ligthModel.check_positive_flux_profile(kwargs_list)
        assert not bool


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            lighModel = LightModel(light_model_list=["WRONG"])
        with self.assertRaises(ValueError):
            lighModel = LightModel(light_model_list=["UNIFORM"])
            lighModel.light_3d(r=1, kwargs_list=[{"amp": 1}])
        with self.assertRaises(ValueError):
            lighModel = LightModel(light_model_list=["UNIFORM"])
            lighModel.profile_type_list = ["WRONG"]
            lighModel.functions_split(x=0, y=0, kwargs_list=[{}])
        with self.assertRaises(ValueError):
            lighModel = LightModel(light_model_list=["UNIFORM"])
            lighModel.profile_type_list = ["WRONG"]
            lighModel.num_param_linear(kwargs_list=[{}])
        with self.assertRaises(ValueError):
            lighModel = LightModel(light_model_list=["UNIFORM"])
            lighModel.profile_type_list = ["WRONG"]
            lighModel.update_linear(param=[1], i=0, kwargs_list=[{}])
        with self.assertRaises(ValueError):
            lighModel = LightModel(light_model_list=["UNIFORM"])
            lighModel.profile_type_list = ["WRONG"]
            lighModel.total_flux(kwargs_list=[{}])


if __name__ == "__main__":
    pytest.main()



import astropy


if float(astropy.__version__[0]) < 5.0:
    from astropy.cosmology.core import isiterable

    DeprecationWarning(
        "Astropy<5 is going to be deprecated soon. This is in combination with Python version<3.8."
        "We recommend you to update astropy to the latest versionbut keep supporting your settings for "
        "the time being."
    )
else:
    from astropy.cosmology.utils import isiterable
#
from astropy import units
import numpy as np
import copy
from scipy.interpolate import interp1d


class CosmoInterp(object):
    """Class which interpolates the comoving transfer distance and then computes angular
    diameter distances from it This class is modifying the astropy.cosmology
    routines."""

    def __init__(
        self,
        cosmo=None,
        z_stop=None,
        num_interp=None,
        ang_dist_list=None,
        z_list=None,
        Ok0=None,
        K=None,
    ):
        """

        :param cosmo: astropy.cosmology instance (version 4.0 as private functions need to be supported)
        :param z_stop: maximum redshift for the interpolation
        :param num_interp: int, number of interpolating steps
        :param ang_dist_list: array of angular diameter distances in Mpc to be interpolated (optional)
        :param z_list: list of redshifts corresponding to ang_dist_list (optional)
        :param Ok0: Omega_k(z=0)
        :param K: Omega_k / (hubble distance)^2  in Mpc^-2
        """
        if cosmo is None:
            if Ok0 is None:
                Ok0 = 0
                K = 0
            self.Ok0 = Ok0
            self.k = K / units.Mpc**2  # in units inverse Mpc^2
            self._comoving_distance_interpolation_func = self._interpolate_ang_dist(
                ang_dist_list, z_list, self.Ok0, self.k.value
            )

        else:
            self._cosmo = cosmo
            self.Ok0 = self._cosmo._Ok0
            dh = self._cosmo._hubble_distance
            self.k = -self.Ok0 / dh**2

            if float(astropy.__version__[0]) < 5.0:
                from lenstronomy.Cosmo._cosmo_interp_astropy_v4 import (
                    CosmoInterp as CosmoInterp_,
                )

                self._comoving_interp = CosmoInterp_(cosmo)
            else:
                from lenstronomy.Cosmo._cosmo_interp_astropy_v5 import (
                    CosmoInterp as CosmoInterp_,
                )

                self._comoving_interp = CosmoInterp_(cosmo)
            self._comoving_distance_interpolation_func = (
                self._interpolate_comoving_distance(
                    z_start=0, z_stop=z_stop, num_interp=num_interp
                )
            )
        self._abs_sqrt_k = np.sqrt(abs(self.k))

    def _comoving_distance_interp(self, z):
        """

        :param z: redshift to which the comoving distance is calculated
        :return: comoving distance in units Mpc
        """
        return self._comoving_distance_interpolation_func(z) * units.Mpc

    def angular_diameter_distance(self, z):
        """Angular diameter distance in Mpc at a given redshift.

        This gives the proper (sometimes called 'physical') transverse
        distance corresponding to an angle of 1 radian for an object
        at redshift ``z``.

        Weinberg, 1972, pp 421-424; Weedman, 1986, pp 65-67; Peebles,
        1993, pp 325-327.

        Parameters
        ----------
        z : array_like
          Input redshifts.  Must be 1D or scalar.

        Returns
        -------
        d : `~astropy.units.Quantity`
          Angular diameter distance in Mpc at each input redshift.
        """

        if isiterable(z):
            z = np.asarray(z)

        return self.comoving_transverse_distance(z) / (1.0 + z)

    def angular_diameter_distance_z1z2(self, z1, z2):
        """Angular diameter distance between objects at 2 redshifts. Useful for
        gravitational lensing.

        Parameters
        ----------
        z1, z2 : array_like, shape (N,)
          Input redshifts. z2 must be large than z1.

        Returns
        -------
        d : `~astropy.units.Quantity`, shape (N,) or single if input scalar
          The angular diameter distance between each input redshift
          pair.
        """

        z1 = np.asanyarray(z1)
        z2 = np.asanyarray(z2)
        return self._comoving_transverse_distance_z1z2(z1, z2) / (1.0 + z2)

    def comoving_transverse_distance(self, z):
        """Comoving transverse distance in Mpc at a given redshift.

        This value is the transverse comoving distance at redshift ``z``
        corresponding to an angular separation of 1 radian. This is
        the same as the comoving distance if omega_k is zero (as in
        the current concordance lambda CDM model).

        Parameters
        ----------
        z : array_like
          Input redshifts.  Must be 1D or scalar.

        Returns
        -------
        d : `~astropy.units.Quantity`
          Comoving transverse distance in Mpc at each input redshift.

        Notes
        -----
        This quantity also called the 'proper motion distance' in some
        texts.
        """

        return self._comoving_transverse_distance_z1z2(0, z)

    def _comoving_transverse_distance_z1z2(self, z1, z2):
        """Comoving transverse distance in Mpc between two redshifts.

        This value is the transverse comoving distance at redshift
        ``z2`` as seen from redshift ``z1`` corresponding to an
        angular separation of 1 radian. This is the same as the
        comoving distance if omega_k is zero (as in the current
        concordance lambda CDM model).

        Parameters
        ----------
        z1, z2 : array_like, shape (N,)
          Input redshifts.  Must be 1D or scalar.

        Returns
        -------
        d : `~astropy.units.Quantity`
          Comoving transverse distance in Mpc between input redshift.

        Notes
        -----
        This quantity is also called the 'proper motion distance' in
        some texts.
        """

        dc = self._comoving_distance_z1z2(z1, z2)
        if np.fabs(self.Ok0) < 1.0e-6:
            return dc
        elif self.k < 0:
            return 1.0 / self._abs_sqrt_k * np.sinh(self._abs_sqrt_k.value * dc.value)
        else:
            return 1.0 / self._abs_sqrt_k * np.sin(self._abs_sqrt_k.value * dc.value)

    def _comoving_distance_z1z2(self, z1, z2):
        """Comoving line-of-sight distance in Mpc between objects at redshifts z1 and
        z2.

        The comoving distance along the line-of-sight between two
        objects remains constant with time for objects in the Hubble
        flow.

        Parameters
        ----------
        z1, z2 : array_like, shape (N,)
          Input redshifts.  Must be 1D or scalar.

        Returns
        -------
        d : `~astropy.units.Quantity`
          Comoving distance in Mpc between each input redshift.
        """
        return self._comoving_distance_interp(z2) - self._comoving_distance_interp(z1)

    def _interpolate_comoving_distance(self, z_start, z_stop, num_interp):
        """Interpolates the comoving distance.

        :param z_start: starting redshift range (should be zero)
        :param z_stop: highest redshift to which to compute the comoving distance
        :param num_interp: number of steps uniformly spread in redshift
        :return: interpolation object in this class
        """
        z_steps = np.linspace(start=z_start, stop=z_stop, num=num_interp + 1)
        running_dist = 0
        ang_dist = np.zeros(num_interp + 1)
        for i in range(num_interp):
            delta_dist = self._comoving_interp._integral_comoving_distance_z1z2(
                z_steps[i], z_steps[i + 1]
            )
            running_dist += delta_dist.value
            ang_dist[i + 1] = copy.deepcopy(running_dist)
        return interp1d(z_steps, ang_dist)

    def _interpolate_ang_dist(self, ang_dist_list, z_list, Ok0, K):
        """Translates angular diameter distances to transversal comoving distances.

        :param ang_dist_list: angular diameter distances in units Mpc
        :type ang_dist_list: numpy array
        :param z_list: redshifts corresponding to ang_dist_list
        :type z_list: numpy array
        :param Ok0: Omega_k(z=0)
        :param K: Omega_k / (hubble distance)^2 in Mpc^-2
        :return: interpolation function of transversal comoving diameter distance [Mpc]
        """
        ang_dist_list = np.asanyarray(ang_dist_list)
        z_list = np.asanyarray(z_list)
        if z_list[0] > 0:  # if redshift zero is not in input, add it
            z_list = np.append(0, z_list)
            ang_dist_list = np.append(0, ang_dist_list)
        if np.fabs(Ok0) < 1.0e-6:
            comoving_dist_list = ang_dist_list * (1.0 + z_list)
        elif K < 0:
            comoving_dist_list = np.arcsinh(
                ang_dist_list * (1.0 + z_list) * np.sqrt(-K)
            ) / np.sqrt(-K)
        else:
            comoving_dist_list = np.arcsin(
                ang_dist_list * (1.0 + z_list) * np.sqrt(K)
            ) / np.sqrt(K)
        return interp1d(z_list, comoving_dist_list)



from lenstronomy.ImSim.image_linear_solve import ImageLinearFit
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
from lenstronomy.Util import class_creator

__all__ = ["SingleBandMultiModel"]


class SingleBandMultiModel(ImageLinearFit):
    """Class to simulate/reconstruct images in multi-band option. This class calls
    functions of image_model.py with different bands with decoupled linear parameters
    and the option to pass/select different light models for the different bands.

    the class supports keyword arguments 'index_lens_model_list', 'index_source_light_model_list',
    'index_lens_light_model_list', 'index_point_source_model_list', 'index_optical_depth_model_list' in kwargs_model
    These arguments should be lists of length the number of imaging bands available and each entry in the list
    is a list of integers specifying the model components being evaluated for the specific band.

    E.g. there are two bands, and you want to different light profiles being modeled.
    - you define two different light profiles lens_light_model_list = ['SERSIC', 'SERSIC']
    - set index_lens_light_model_list = [[0], [1]]
    - (optional) for now all the parameters between the two light profiles are independent in the model. You have
    the possibility to join a subset of model parameters (e.g. joint centroid). See the Param() class for documentation.
    """

    def __init__(
        self,
        multi_band_list,
        kwargs_model,
        likelihood_mask_list=None,
        band_index=0,
        kwargs_pixelbased=None,
        linear_solver=True,
    ):
        """

        :param multi_band_list: list of imaging band configurations [[kwargs_data, kwargs_psf, kwargs_numerics],[...], ...]
        :param kwargs_model: model option keyword arguments
        :param likelihood_mask_list: list of likelihood masks (booleans with size of the individual images
        :param band_index: integer, index of the imaging band to model
        :param kwargs_pixelbased: keyword arguments with various settings related to the pixel-based solver
         (see SLITronomy documentation)
        :param linear_solver: bool, if True (default) fixes the linear amplitude parameters 'amp' (avoid sampling) such
         that they get overwritten by the linear solver solution.
        """
        self.type = "single-band-multi-model"
        if likelihood_mask_list is None:
            likelihood_mask_list = [None for _ in range(len(multi_band_list))]
        (
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            extinction_class,
        ) = class_creator.create_class_instances(band_index=band_index, **kwargs_model)
        kwargs_data = multi_band_list[band_index][0]
        kwargs_psf = multi_band_list[band_index][1]
        kwargs_numerics = multi_band_list[band_index][2]
        data_i = ImageData(**kwargs_data)
        psf_i = PSF(**kwargs_psf)

        index_lens_model_list = kwargs_model.get(
            "index_lens_model_list", [None for _ in range(len(multi_band_list))]
        )
        self._index_lens_model = index_lens_model_list[band_index]
        index_source_list = kwargs_model.get(
            "index_source_light_model_list", [None for _ in range(len(multi_band_list))]
        )
        self._index_source = index_source_list[band_index]
        index_lens_light_list = kwargs_model.get(
            "index_lens_light_model_list", [None for _ in range(len(multi_band_list))]
        )
        self._index_lens_light = index_lens_light_list[band_index]
        index_point_source_list = kwargs_model.get(
            "index_point_source_model_list", [None for _ in range(len(multi_band_list))]
        )
        self._index_point_source = index_point_source_list[band_index]
        index_optical_depth = kwargs_model.get(
            "index_optical_depth_model_list",
            [None for _ in range(len(multi_band_list))],
        )
        self._index_optical_depth = index_optical_depth[band_index]

        super(SingleBandMultiModel, self).__init__(
            data_i,
            psf_i,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            extinction_class,
            kwargs_numerics=kwargs_numerics,
            likelihood_mask=likelihood_mask_list[band_index],
            kwargs_pixelbased=kwargs_pixelbased,
            linear_solver=linear_solver,
        )

    def image(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        unconvolved=False,
        source_add=True,
        lens_light_add=True,
        point_source_add=True,
    ):
        """Make an image with a realisation of linear parameter values "param".

        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param kwargs_ps: keyword arguments corresponding to "other" parameters, such as
            external shear and point source image positions
        :param unconvolved: if True: returns the unconvolved light distribution (prefect
            seeing)
        :param source_add: if True, compute source, otherwise without
        :param lens_light_add: if True, compute lens light, otherwise without
        :param point_source_add: if True, add point sources, otherwise without
        :return: 2d array of surface brightness pixels of the simulation
        """
        (
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
        ) = self.select_kwargs(
            kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps, kwargs_extinction
        )
        return self._image(
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
            kwargs_special=kwargs_special,
            unconvolved=unconvolved,
            source_add=source_add,
            lens_light_add=lens_light_add,
            point_source_add=point_source_add,
        )

    def source_surface_brightness(
        self,
        kwargs_source,
        kwargs_lens=None,
        kwargs_extinction=None,
        kwargs_special=None,
        unconvolved=False,
        de_lensed=False,
        k=None,
        update_pixelbased_mapping=True,
    ):
        """Computes the source surface brightness distribution.

        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_extinction: list of keyword arguments of extinction model
        :param unconvolved: if True: returns the unconvolved light distribution (prefect
            seeing)
        :param de_lensed: if True: returns the un-lensed source surface brightness
            profile, otherwise the lensed.
        :param k: integer, if set, will only return the model of the specific index
        :return: 2d array of surface brightness pixels
        """
        kwargs_lens_i, kwargs_source_i, _, _, kwargs_extinction_i = self.select_kwargs(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light=None,
            kwargs_ps=None,
            kwargs_extinction=kwargs_extinction,
        )
        return self._source_surface_brightness(
            kwargs_source_i,
            kwargs_lens_i,
            kwargs_extinction=kwargs_extinction_i,
            kwargs_special=kwargs_special,
            unconvolved=unconvolved,
            de_lensed=de_lensed,
            k=k,
            update_pixelbased_mapping=update_pixelbased_mapping,
        )

    def lens_surface_brightness(self, kwargs_lens_light, unconvolved=False, k=None):
        """Computes the lens surface brightness distribution.

        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param unconvolved: if True, returns unconvolved surface brightness (perfect
            seeing), otherwise convolved with PSF kernel
        :return: 2d array of surface brightness pixels
        """
        _, _, kwargs_lens_light_i, kwargs_ps_i, _ = self.select_kwargs(
            kwargs_lens=None,
            kwargs_source=None,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_ps=None,
            kwargs_extinction=None,
        )
        return self._lens_surface_brightness(
            kwargs_lens_light_i, unconvolved=unconvolved, k=k
        )

    def point_source(
        self,
        kwargs_ps,
        kwargs_lens=None,
        kwargs_special=None,
        unconvolved=False,
        k=None,
    ):
        """Computes the point source positions and paints PSF convolutions on them.

        :param kwargs_ps:
        :param kwargs_lens:
        :param kwargs_special:
        :param unconvolved:
        :param k:
        :return:
        """
        kwargs_lens_i, _, _, kwargs_ps_i, _ = self.select_kwargs(
            kwargs_lens=kwargs_lens,
            kwargs_source=None,
            kwargs_lens_light=None,
            kwargs_ps=kwargs_ps,
            kwargs_extinction=None,
        )
        return self._point_source(
            kwargs_ps=kwargs_ps_i,
            kwargs_lens=kwargs_lens_i,
            kwargs_special=kwargs_special,
            unconvolved=unconvolved,
            k=k,
        )

    def image_linear_solve(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        inv_bool=False,
    ):
        """Computes the image (lens and source surface brightness with a given lens
        model).

        The linear parameters are computed with a weighted linear least square
        optimization (i.e. flux normalization of the brightness profiles)

        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param kwargs_ps: keyword arguments corresponding to "other" parameters, such as
            external shear and point source image positions
        :param inv_bool: if True, invert the full linear solver Matrix Ax = y for the
            purpose of the covariance matrix.
        :return: 1d array of surface brightness pixels of the optimal solution of the
            linear parameters to match the data
        """
        (
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
        ) = self.select_kwargs(
            kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps, kwargs_extinction
        )
        wls_model, error_map, cov_param, param = self._image_linear_solve(
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
            kwargs_special,
            inv_bool=inv_bool,
        )
        # For the interfometric likelihood method,
        # return the array2 of [array1, array2] of the model output of _image_linear_solver.
        if self.Data.likelihood_method() == "interferometry_natwt":
            wls_model = wls_model[1]
        return wls_model, error_map, cov_param, param

    def likelihood_data_given_model(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        source_marg=False,
        linear_prior=None,
        check_positive_flux=False,
        linear_solver=True,
    ):
        """Computes the likelihood of the data given a model This is specified with the
        non-linear parameters and a linear inversion and prior marginalisation.

        :param kwargs_lens:
        :param kwargs_source:
        :param kwargs_lens_light:
        :param kwargs_ps:
        :param check_positive_flux: bool, if True, checks whether the linear inversion
            resulted in non-negative flux components and applies a punishment in the
            likelihood if so.
        :return: log likelihood (natural logarithm) (sum of the log likelihoods of the
            individual images)
        """
        # generate image
        (
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
        ) = self.select_kwargs(
            kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps, kwargs_extinction
        )
        return self._likelihood_data_given_model(
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
            kwargs_special,
            source_marg=source_marg,
            linear_prior=linear_prior,
            check_positive_flux=check_positive_flux,
            linear_solver=self._linear_solver,
        )

    def num_param_linear(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
    ):
        """

        :return: number of linear coefficients to be solved for in the linear inversion
        """
        if self._linear_solver is False:
            return 0
        (
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
        ) = self.select_kwargs(kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps)
        num = self._num_param_linear(
            kwargs_lens_i, kwargs_source_i, kwargs_lens_light_i, kwargs_ps_i
        )
        return num

    def linear_response_matrix(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
    ):
        """Computes the linear response matrix (m x n), with n beeing the data size and
        m being the coefficients.

        :param kwargs_lens:
        :param kwargs_source:
        :param kwargs_lens_light:
        :param kwargs_ps:
        :return:
        """
        (
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
        ) = self.select_kwargs(
            kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps, kwargs_extinction
        )
        A = self._linear_response_matrix(
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
            kwargs_special,
        )
        return A

    def error_map_source(
        self, kwargs_source, x_grid, y_grid, cov_param, model_index_select=True
    ):
        """Variance of the linear source reconstruction in the source plane coordinates,
        computed by the diagonal elements of the covariance matrix of the source
        reconstruction as a sum of the errors of the basis set.

        :param kwargs_source: keyword arguments of source model
        :param x_grid: x-axis of positions to compute error map
        :param y_grid: y-axis of positions to compute error map
        :param cov_param: covariance matrix of liner inversion parameters
        :param model_index_select: boolean, if True, selects the model components of
            this band (default). If False, assumes input kwargs_source is already
            selected list.
        :return: diagonal covariance errors at the positions (x_grid, y_grid)
        """
        if self._index_source is None or model_index_select is False:
            kwargs_source_i = kwargs_source
        else:
            kwargs_source_i = [kwargs_source[k] for k in self._index_source]
        return self._error_map_source(kwargs_source_i, x_grid, y_grid, cov_param)

    def error_response(self, kwargs_lens, kwargs_ps, kwargs_special):
        """Returns the 1d array of the error estimate corresponding to the data
        response.

        :return: 1d numpy array of response, 2d array of additional errors (e.g. point
            source uncertainties)
        """
        (
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
        ) = self.select_kwargs(
            kwargs_lens,
            kwargs_source=None,
            kwargs_lens_light=None,
            kwargs_ps=kwargs_ps,
            kwargs_extinction=None,
        )
        return self._error_response(
            kwargs_lens_i, kwargs_ps_i, kwargs_special=kwargs_special
        )

    def update_linear_kwargs(
        self, param, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
    ):
        """

        links linear parameters to kwargs arguments
        ATTENTION: this function requires input dictionary lists to be already contracted to the ones applied to the
        specific band

        :param param: linear parameter vector corresponding to the response matrix
        :return: updated list of kwargs with linear parameter values
        """
        (
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
        ) = self.select_kwargs(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
            kwargs_extinction=None,
        )
        return self._update_linear_kwargs(
            param, kwargs_lens_i, kwargs_source_i, kwargs_lens_light_i, kwargs_ps_i
        )

    def extinction_map(self, kwargs_extinction=None, kwargs_special=None):
        """Differential extinction per pixel.

        :param kwargs_extinction: list of keyword arguments corresponding to the optical
            depth models tau, such that extinction is exp(-tau)
        :param kwargs_special: keyword arguments, additional parameter to the extinction
        :return: 2d array of size of the image
        """
        _, _, _, _, kwargs_extinction_i = self.select_kwargs(
            kwargs_extinction=kwargs_extinction
        )
        return self._extinction_map(kwargs_extinction_i, kwargs_special)

    def linear_param_from_kwargs(self, kwargs_source, kwargs_lens_light, kwargs_ps):
        """Inverse function of update_linear() returning the linear amplitude list for
        the keyword argument list.

        :param kwargs_source:
        :param kwargs_lens_light:
        :param kwargs_ps:
        :return: list of linear coefficients
        """
        _, kwargs_source_i, kwargs_lens_light_i, kwargs_ps_i, _ = self.select_kwargs(
            kwargs_lens=None,
            kwargs_source=kwargs_source,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_ps=kwargs_ps,
            kwargs_extinction=None,
        )
        return self._linear_param_from_kwargs(
            kwargs_source_i, kwargs_lens_light_i, kwargs_ps_i
        )

    def select_kwargs(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
    ):
        """Select subset of kwargs lists referenced to this imaging band.

        :param kwargs_lens:
        :param kwargs_source:
        :param kwargs_lens_light:
        :param kwargs_ps:
        :return:
        """
        if self._index_lens_model is None or kwargs_lens is None:
            kwargs_lens_i = kwargs_lens
        else:
            kwargs_lens_i = [kwargs_lens[k] for k in self._index_lens_model]
        if self._index_source is None or kwargs_source is None:
            kwargs_source_i = kwargs_source
        else:
            kwargs_source_i = [kwargs_source[k] for k in self._index_source]
        if self._index_lens_light is None or kwargs_lens_light is None:
            kwargs_lens_light_i = kwargs_lens_light
        else:
            kwargs_lens_light_i = [kwargs_lens_light[k] for k in self._index_lens_light]
        if self._index_point_source is None or kwargs_ps is None:
            kwargs_ps_i = kwargs_ps
        else:
            kwargs_ps_i = [kwargs_ps[k] for k in self._index_point_source]
        if self._index_optical_depth is None or kwargs_extinction is None:
            kwargs_extinction_i = kwargs_extinction
        else:
            kwargs_extinction_i = [
                kwargs_extinction[k] for k in self._index_optical_depth
            ]
        return (
            kwargs_lens_i,
            kwargs_source_i,
            kwargs_lens_light_i,
            kwargs_ps_i,
            kwargs_extinction_i,
        )



import unittest
from lenstronomy.SimulationAPI.ObservationConfig.Roman import Roman
from lenstronomy.SimulationAPI.observation_api import Instrument, SingleBand
import lenstronomy.Util.util as util
import os
import astropy.io.fits as pyfits


class TestRoman(unittest.TestCase):
    def setUp(self):
        self.F062 = Roman()  # default is F062
        self.F087 = Roman(band="F087", survey_mode="microlensing")
        self.F106 = Roman(band="F106", psf_type="GAUSSIAN")
        self.F129 = Roman(band="F129", psf_type="GAUSSIAN")
        self.F158 = Roman(band="F158", psf_type="GAUSSIAN")
        self.F184 = Roman(band="F184", psf_type="GAUSSIAN")
        self.F146 = Roman(band="F146", survey_mode="microlensing", psf_type="GAUSSIAN")

        kwargs_F062 = self.F062.kwargs_single_band()
        kwargs_F087 = self.F087.kwargs_single_band()
        kwargs_F106 = self.F106.kwargs_single_band()
        kwargs_F129 = self.F129.kwargs_single_band()
        kwargs_F158 = self.F158.kwargs_single_band()
        kwargs_F184 = self.F184.kwargs_single_band()
        kwargs_F146 = self.F146.kwargs_single_band()

        self.F062_band = SingleBand(**kwargs_F062)
        self.F087_band = SingleBand(**kwargs_F087)
        self.F106_band = SingleBand(**kwargs_F106)
        self.F129_band = SingleBand(**kwargs_F129)
        self.F158_band = SingleBand(**kwargs_F158)
        self.F184_band = SingleBand(**kwargs_F184)
        self.F146_band = SingleBand(**kwargs_F146)

        # dictionaries mapping Roman kwargs to SingleBand kwargs
        self.camera_settings = {
            "read_noise": "_read_noise",
            "pixel_scale": "pixel_scale",
            "ccd_gain": "ccd_gain",
        }
        self.obs_settings = {
            "exposure_time": "_exposure_time",
            "sky_brightness": "_sky_brightness_",
            "magnitude_zero_point": "_magnitude_zero_point",
            "num_exposures": "_num_exposures",
            "seeing": "_seeing",
            "psf_type": "_psf_type",
        }

        self.instrument = Instrument(**self.F062.camera)

    def test_Roman_class(self):
        default = self.F062
        explicit_F062 = Roman(band="F062")
        self.assertEqual(explicit_F062.camera, default.camera)
        self.assertEqual(explicit_F062.obs, default.obs)

        with self.assertRaises(ValueError):
            bad_band = Roman(band="g")

        with self.assertRaises(ValueError):
            bad_band_2 = Roman(band="9")

        with self.assertRaises(ValueError):
            bad_psf = Roman(psf_type="blah")

        with self.assertRaises(ValueError):
            bad_band_wide = Roman(band="F087")

        with self.assertRaises(ValueError):
            bad_band_microlensing = Roman(band="F062", survey_mode="microlensing")

        with self.assertRaises(ValueError):
            bad_survey_mode = Roman(survey_mode="blah")

    def test_Roman_camera(self):
        # comparing camera settings in Roman instance with those in Instrument instance
        for config, setting in self.camera_settings.items():
            self.assertEqual(
                self.F062.camera[config],
                getattr(self.instrument, setting),
                msg=f"{config} did not match",
            )

    def test_Roman_obs(self):
        # comparing obs settings in HST instance with those in SingleBand instance
        for config, setting in self.obs_settings.items():
            self.assertEqual(
                self.F062.obs[config],
                getattr(self.F062_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.F087.obs[config],
                getattr(self.F087_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.F106.obs[config],
                getattr(self.F106_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.F129.obs[config],
                getattr(self.F129_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.F158.obs[config],
                getattr(self.F158_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.F184.obs[config],
                getattr(self.F184_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.F146.obs[config],
                getattr(self.F146_band, setting),
                msg=f"{config} did not match",
            )

    def test_Roman_psf_pixel(self):
        self.F062_pixel = Roman(psf_type="PIXEL")
        import lenstronomy

        module_path = os.path.dirname(lenstronomy.__file__)
        psf_filename = os.path.join(
            module_path, "SimulationAPI/ObservationConfig/PSF_models/F062.fits"
        )
        kernel = pyfits.getdata(psf_filename)
        self.assertEqual(
            self.F062_pixel.obs["kernel_point_source"].all(),
            kernel.all(),
            msg="PSF did not match",
        )

    def test_kwargs_single_band(self):
        kwargs_F062 = util.merge_dicts(self.F062.camera, self.F062.obs)
        self.assertEqual(self.F062.kwargs_single_band(), kwargs_F062)


if __name__ == "__main__":
    unittest.main()



import pytest
import numpy.testing as npt
import numpy as np
from lenstronomy.GalKin import velocity_util


class TestVelocityUtil(object):
    """"""

    def setup_method(self):
        pass

    def test_sample_gaussian(self):
        np.random.seed(41)
        n = 1000
        FWHM = 1
        pos_x, pos_y = [], []
        for i in range(n):
            x, y = velocity_util.displace_PSF_gaussian(0, 0, FWHM)
            pos_x.append(x)
            pos_y.append(y)

        sigma = FWHM / (2 * np.sqrt(2 * np.log(2)))
        sigma_x = np.std(pos_x)
        npt.assert_almost_equal(sigma, sigma_x, decimal=1)

    def test_sample_moffat(self):
        np.random.seed(41)
        n = 10000
        FWHM = 1
        beta = 2.6
        r_list = []
        for i in range(n):
            r = velocity_util.draw_moffat_r(FWHM, beta)
            r_list.append(r)
        x_array = np.linspace(0, 4 * FWHM, num=100)
        r_hist, bins = np.histogram(r_list, bins=x_array, density=True)
        alpha = velocity_util.moffat_fwhm_alpha(FWHM, beta)

        x_ = x_array[1:] - x_array[1] + x_array[0]
        f_moffat = velocity_util.moffat_r(x_, alpha=alpha, beta=beta) * x_
        # import matplotlib.pyplot as plt
        # plt.plot(x_, f_moffat, label='moffat')
        # plt.plot(x_, r_hist, label='sampled')
        # plt.legend()
        # plt.show()
        npt.assert_almost_equal(r_hist, f_moffat, decimal=1)

    def test_displace_PSF_moffat(self):
        FWHM = 1
        beta = 2.6
        np.random.seed(41)
        x, y = 0, 0
        x_d, y_d = velocity_util.displace_PSF_moffat(x, y, FWHM, beta)
        assert x_d != x
        assert y_d != y

    def test_project_2d_random(self):
        r = 1
        R, x, y = velocity_util.project2d_random(r=r)
        assert R <= r

        num = 100000
        r = np.ones(num)
        R, x, y = velocity_util.project2d_random(r=r)
        assert len(R) == num
        # import matplotlib.pyplot as plt
        # bins = np.linspace(0., 1, 100)
        # hist, bins_hist = np.histogram(R, bins=bins, density=True)
        # bins_plot = (bins_hist[1:] + bins_hist[:-1]) / 2.
        # plt.plot(bins_plot, hist, label='regular')
        # plt.legend()
        # plt.show()


if __name__ == "__main__":
    pytest.main()



__author__ = "Matt Gomer"

import numpy as np
from scipy.interpolate import RectBivariateSpline
import matplotlib.pyplot as plt


class KinNNImageAlign(object):
    """Class to rotate and interpolate SKiNN image (which is built along x-axis) to be
    on the grid of the spectral data (e.g., MUSE), rotated to match light distribution
    as defined by the lens model main function is interp_image() which will output a 2D
    image interpolated on the spectra grid."""

    def __init__(self, spectra_inputs, imaging_inputs, kin_nn_inputs):
        """Initialize input data.

        :param spectra_inputs: dictionary which encodes grid and transformation
            information for kinematic data :'image': contains 2d image used to calculate
            grid coordinates :'transform_pix2angle': transformation matrix to convert
            from pixel xy to ra/dec :'ra_at_xy0': ra coordinate at pixel (0,0)
            :'dec_at_xy0': dec coordinate at pixel (0,0)
        :param imaging_inputs: dictionary which encodes grid and transformation
            information for imaging data :'image': contains 2d image used to calculate
            grid coordinates :'transform_pix2angle': transformation matrix to convert
            from pixel xy to ra/dec :'ra_at_xy0': ra coordinate at pixel (0,0)
            :'dec_at_xy0': dec coordinate at pixel (0,0) :'ellipse_PA': position angle
            of ellipse axis relative to x direction :'offset_x': how many pixels to
            offset the center of the grid to match the kinNN center (x-direction)
            :'offset_y': how many pixels to offset the center of the grid to match the
            kinNN center (y-direction)
        :param kin_nn_inputs: dictionary which encodes grid information for NN output
            data :'image': contains 2d image used to calculate grid coordinates
            :'deltaPix': pixel size
        """
        self.spectra_data = spectra_inputs
        self.imaging_data = imaging_inputs
        self.imaging_deltapix = np.sqrt(
            np.abs(np.linalg.det(self.imaging_data["transform_pix2angle"]))
        )
        self.kinNN_data = kin_nn_inputs
        self.write_npix()

    def update(
        self,
        spectra_inputs=None,
        imaging_inputs=None,
        kin_nn_inputs=None,
        update_npix=False,
    ):
        """Update with inputs."""
        if spectra_inputs is not None:
            self.spectra_data = spectra_inputs
        if imaging_inputs is not None:
            self.imaging_data = imaging_inputs
        if kin_nn_inputs is not None:
            self.kinNN_data = kin_nn_inputs
        if update_npix:
            self.write_npix()

    def write_npix(self):
        """Check that images are squared and write the keyword 'npix'."""
        for input_set in [self.spectra_data, self.imaging_data, self.kinNN_data]:
            # make sure each image is square and add npix to each dictionary
            if "image" in input_set.keys():
                if np.shape(input_set["image"])[0] != np.shape(input_set["image"])[1]:
                    raise ValueError("current version only works for square images")
                npix = np.shape(input_set["image"])[0]
                input_set["npix"] = npix

    def pix_coords(self, input_set, flatten=True):
        """Simple function to give pixel coordinates of grid.

        :param input_set: dictionary from above (e.g. spectra_inputs) which completely
            describes grid transformation :boolean flatten: default True; if True,
            return 1D flattened output, if False, return 2D grid :return pixel
            coordinates of grid
        """
        x_grid = np.tile(np.arange(input_set["npix"]), input_set["npix"])
        y_grid = np.repeat(np.arange(input_set["npix"]), input_set["npix"])
        if flatten is True:
            return x_grid, y_grid
        else:
            return x_grid.reshape(input_set["npix"], input_set["npix"]), y_grid.reshape(
                input_set["npix"], input_set["npix"]
            )

    def radec_to_xy(self, ra, dec, xy_to_radec_matrix, ra_atxy0, dec_atxy0):
        """Converts from radec to pixel coordinates.

        :param ra: ra coordinate to transform
        :param dec: dec coordinate to transform
        :param xy_to_radec_matrix: transformation matrix to convert from pixel xy to
            ra/dec
        :param ra_atxy0: ra coordinate at pixel (0,0)
        :param dec_atxy0: dec coordinate at pixel (0,0)
        :return: x and y coordinates
        """
        ra = ra - ra_atxy0
        dec = dec - dec_atxy0
        x, y = np.linalg.inv(xy_to_radec_matrix).dot(np.array([ra, dec]))
        return x, y

    def xy_to_radec(self, x, y, xy_to_radec_matrix, ra_atxy0, dec_atxy0):
        """Converts from pixel coordinates to radec.

        :param x: x coordinate to transform
        :param y: y coordinate to transform
        :param xy_to_radec_matrix: transformation matrix to convert from pixel xy to
            ra/dec
        :param ra_atxy0: ra coordinate at pixel (0,0)
        :param dec_atxy0: dec coordinate at pixel (0,0)
        :return: ra and dec coordinates
        """
        ra, dec = xy_to_radec_matrix.dot(np.array([x, y]))
        return ra + ra_atxy0, dec + dec_atxy0

    def rotate_imaging_into_kin_nn(
        self,
        imaging_x,
        imaging_y,
        ellipse_pa_to_imagingx_angle,
        deltapix_imaging,
        deltapix_kin_nn,
        npix_imaging,
        npix_kin_nn,
        offsetx=0,
        offsety=0,
    ):
        """Rotates and rescales from the x,y imaging coordinate system into the NN
        coordinate system.

        :param imaging_x: imaging x coordinate to transform
        :param imaging_y: imaging y coordinate to transform
        :param ellipse_pa_to_imagingx_angle: (radians) position angle of ellipse major
            axis relative to x in the imaging coordinate system
        :param deltapix_imaging: pixel size of imaging image
        :param deltapix_kin_nn: pixel size of NN image
        :param npix_imaging: number of pixels on a side of the imaging image
        :param npix_kin_nn: number of pixels on a side of the NN image
        :param offsetx: how many pixels to offset the center of the grid to match the
            kinNN center (x-direction)
        :param offsety: how many pixels to offset the center of the grid to match the
            kinNN center (y-direction)
        :return: x and y coordinates in NN coordinate system
        """
        # define rotation matrix to rotate back into alignment
        counterrotation = -ellipse_pa_to_imagingx_angle
        cd1_1 = np.cos(counterrotation)
        cd1_2 = -np.sin(counterrotation)
        cd2_1 = np.sin(counterrotation)
        cd2_2 = np.cos(counterrotation)
        # rotation matrix, applied to matching centers ()
        rotation_by_ellipse_angle = np.array([[cd1_1, cd1_2], [cd2_1, cd2_2]]) * (
            deltapix_imaging / deltapix_kin_nn
        )

        (
            kin_nn_x_at_imagingcenter,
            kin_nn_y_at_imagingcenter,
        ) = rotation_by_ellipse_angle.dot(
            np.array(
                [
                    -npix_imaging / 2 - offsetx / deltapix_imaging,
                    -npix_imaging / 2 - offsety / deltapix_imaging,
                ]
            )
        ) + [
            npix_kin_nn / 2,
            npix_kin_nn / 2,
        ]

        kin_nn_x, kin_nn_y = rotation_by_ellipse_angle.dot(
            np.array([imaging_x, imaging_y])
        )
        return (
            kin_nn_x + kin_nn_x_at_imagingcenter,
            kin_nn_y + kin_nn_y_at_imagingcenter,
        )

    def plot_contour_and_grid(self, xcoords, ycoords, orig_image, color, alpha=0.4):
        """Plotting function for visualization of a grid with a single ellipse contour.

        :param xcoords: grid x coordinates, flattened
        :param ycoords: grid y coordinates, flattened
        :param orig_image: original 2D image
        :param color: plotting color
        :param alpha: plotting alpha
        """
        ell_cond_2d = np.isclose(orig_image, 0.1, rtol=5e-02)
        ell_cond = ell_cond_2d.flatten()
        plt.scatter(
            xcoords[ell_cond], ycoords[ell_cond], color=color, alpha=alpha, s=10
        )
        plt.scatter(xcoords, ycoords, color=color, alpha=alpha, s=1)

    def spectragrid_in_radec(self):
        """Calculates ra and dec coordinates of spectra input coordinate grid.

        :return: ra and dec coordinates
        """
        spectra_x, spectra_y = self.pix_coords(self.spectra_data, flatten=True)
        spectra_ra, spectra_dec = self.xy_to_radec(
            spectra_x,
            spectra_y,
            self.spectra_data["transform_pix2angle"],
            self.spectra_data["ra_at_xy0"],
            self.spectra_data["dec_at_xy0"],
        )
        return spectra_ra, spectra_dec

    def spectragrid_in_imagingxy(self):
        """Calculates x and y coordinates in the imaging coordinate system of the
        original spectra input grid.

        :return: x and y coordinates
        """
        spectra_ra, spectra_dec = self.spectragrid_in_radec()
        spectra_coords_in_imaging_x, spectra_coords_in_imaging_y = self.radec_to_xy(
            spectra_ra,
            spectra_dec,
            self.imaging_data["transform_pix2angle"],
            self.imaging_data["ra_at_xy0"],
            self.imaging_data["dec_at_xy0"],
        )
        return spectra_coords_in_imaging_x, spectra_coords_in_imaging_y

    def spectragrid_in_kin_nn_xy(self):
        """Calculates x and y coordinates in the NN coordinate system of the original
        spectra input grid.

        :return: x and y coordinates
        """
        (
            spectra_coords_in_imaging_x,
            spectra_coords_in_imaging_y,
        ) = self.spectragrid_in_imagingxy()
        kin_nn_x, kin_nn_y = self.rotate_imaging_into_kin_nn(
            spectra_coords_in_imaging_x,
            spectra_coords_in_imaging_y,
            self.imaging_data["ellipse_PA"],
            self.imaging_deltapix,
            self.kinNN_data["deltaPix"],
            self.imaging_data["npix"],
            self.kinNN_data["npix"],
            offsetx=self.imaging_data["offset_x"],
            offsety=self.imaging_data["offset_y"],
        )
        return kin_nn_x, kin_nn_y

    def interp_image(self):
        """Interpolates kinNN image at the coordinates of the transformed spectra grid.

        :return: interpolated image which lines up with spectra coordinates
        """
        spectra_kin_nn_x, spectra_kin_nn_y = self.spectragrid_in_kin_nn_xy()
        x_axis = np.arange(self.kinNN_data["npix"])
        y_axis = np.arange(self.kinNN_data["npix"])
        interp_fcn = RectBivariateSpline(x_axis, y_axis, self.kinNN_data["image"])
        # y and x are flipped in RectBivariateSpline call:
        flux_interp = interp_fcn.ev(spectra_kin_nn_y, spectra_kin_nn_x).reshape(
            self.spectra_data["npix"], self.spectra_data["npix"]
        )
        return flux_interp



__author__ = "sibirrer"

import numpy as np

from lenstronomy.ImSim.image_linear_solve import ImageLinearFit
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
import numpy.testing as npt


class TestImageLinearFit(object):
    def setup_method(self):
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 100  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg, inverse=True
        )
        data_class = ImageData(**kwargs_data)
        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        psf_class = PSF(**kwargs_psf)

        kwargs_sis = {"theta_E": 1.0, "center_x": 0, "center_y": 0}

        lens_model_list = ["SIS"]
        self.kwargs_lens = [kwargs_sis]
        lens_model_class = LensModel(lens_model_list=lens_model_list)

        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        phi, q = 0.2, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 7,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_light_model_list = ["SERSIC"]
        self.kwargs_lens_light = [kwargs_sersic]
        lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        self.kwargs_source = [kwargs_sersic_ellipse]
        source_model_class = LightModel(light_model_list=source_model_list)
        self.kwargs_ps = [
            {"ra_source": 0.01, "dec_source": 0.0, "source_amp": 1.0}
        ]  # quasar point source position in the source plane and intrinsic brightness
        point_source_class = PointSource(
            point_source_type_list=["SOURCE_POSITION"], fixed_magnification_list=[True]
        )
        kwargs_numerics = {
            "supersampling_factor": 2,
            "supersampling_convolution": False,
        }

        self.imageModel = ImageLinearFit(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )

    def test_linear_param_from_kwargs(self):
        param = self.imageModel.linear_param_from_kwargs(
            self.kwargs_source, self.kwargs_lens_light, self.kwargs_ps
        )
        assert param[0] == self.kwargs_source[0]["amp"]
        assert param[1] == self.kwargs_lens_light[0]["amp"]
        assert param[2] == self.kwargs_ps[0]["source_amp"]

    def test_update_linear_kwargs(self):
        num = self.imageModel.num_param_linear(
            self.kwargs_lens, self.kwargs_source, self.kwargs_lens_light, self.kwargs_ps
        )
        param = np.ones(num) * 10
        (
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
        ) = self.imageModel.update_linear_kwargs(
            param,
            kwargs_lens=self.kwargs_lens,
            kwargs_source=self.kwargs_source,
            kwargs_lens_light=self.kwargs_lens_light,
            kwargs_ps=self.kwargs_ps,
        )
        assert kwargs_source[0]["amp"] == 10

    def test_error_response(self):
        C_D_response, model_error = self.imageModel.error_response(
            kwargs_lens=self.kwargs_lens, kwargs_ps=self.kwargs_ps, kwargs_special=None
        )
        npt.assert_almost_equal(model_error, 0)



__author__ = "sibirrer"

# this file contains a class to make a Moffat profile

__all__ = ["Moffat"]


class Moffat(object):
    """This class contains functions to evaluate a Moffat surface brightness profile.

    .. math::

        I(r) = I_0 * (1 + (r/\\alpha)^2)^{-\\beta}

    with :math:`I_0 = amp`.
    """

    def __init__(self):
        self.param_names = ["amp", "alpha", "beta", "center_x", "center_y"]
        self.lower_limit_default = {
            "amp": 0,
            "alpha": 0,
            "beta": 0,
            "center_x": -100,
            "center_y": -100,
        }
        self.upper_limit_default = {
            "amp": 100,
            "alpha": 10,
            "beta": 10,
            "center_x": 100,
            "center_y": 100,
        }

    def function(self, x, y, amp, alpha, beta, center_x=0, center_y=0):
        """2D Moffat profile.

        :param x: x-position (angle)
        :param y: y-position (angle)
        :param amp: normalization
        :param alpha: scale
        :param beta: exponent
        :param center_x: x-center
        :param center_y: y-center
        :return: surface brightness
        """

        x_shift = x - center_x
        y_shift = y - center_y
        return amp * (1.0 + (x_shift**2 + y_shift**2) / alpha**2) ** (-beta)



from lenstronomy.Util import image_util
from lenstronomy.Util import kernel_util
import numpy as np

__all__ = ["PointSourceRendering"]


class PointSourceRendering(object):
    """Numerics to compute the point source response on an image."""

    def __init__(self, pixel_grid, supersampling_factor, psf):
        """

        :param pixel_grid: PixelGrid() instance
        :param supersampling_factor: int, factor of supersampling of point source
        :param psf: PSF() instance
        """
        self._pixel_grid = pixel_grid
        self._nx, self._ny = self._pixel_grid.num_pixel_axes
        self._supersampling_factor = supersampling_factor
        self._psf = psf

    def point_source_rendering(self, ra_pos, dec_pos, amp):
        """

        :param ra_pos: list of RA positions of point source(s)
        :param dec_pos: list of DEC positions of point source(s)
        :param amp: list of amplitudes of point source(s)
        :return: 2d numpy array of size of the image with the point source(s) rendered
        """
        subgrid = self._supersampling_factor
        x_pos, y_pos = self._pixel_grid.map_coord2pix(ra_pos, dec_pos)
        # translate coordinates to higher resolution grid
        x_pos_subgird = x_pos * subgrid + (subgrid - 1) / 2.0
        y_pos_subgrid = y_pos * subgrid + (subgrid - 1) / 2.0
        kernel_point_source_subgrid = self._kernel_supersampled
        # initialize grid with higher resolution
        subgrid2d = np.zeros((self._nx * subgrid, self._ny * subgrid))
        # add_layer2image
        if len(x_pos) > len(amp):
            raise ValueError(
                "there are %s images appearing but only %s amplitudes provided!"
                % (len(x_pos), len(amp))
            )
        for i in range(len(x_pos)):
            subgrid2d = image_util.add_layer2image(
                subgrid2d,
                x_pos_subgird[i],
                y_pos_subgrid[i],
                amp[i] * kernel_point_source_subgrid,
            )
        # re-size grid to data resolution
        grid2d = image_util.re_size(subgrid2d, factor=subgrid)
        return grid2d * subgrid**2

    @property
    def _kernel_supersampled(self):
        if not hasattr(self, "_kernel_supersampled_instance"):
            self._kernel_supersampled_instance = (
                self._psf.kernel_point_source_supersampled(
                    self._supersampling_factor, updata_cache=False
                )
            )
        return self._kernel_supersampled_instance

    def psf_error_map(self, ra_pos, dec_pos, amp, data, fix_psf_error_map=False):
        """

        :param ra_pos: image positions of point sources
        :param dec_pos: image positions of point sources
        :param amp: amplitude of modeled point sources
        :param data: 2d numpy array of the data
        :param fix_psf_error_map: bool, if True, estimates the error based on the input (modeled) amplitude, else uses
         the data to do so.
        :return: 2d array of size of the image with error terms (sigma**2) expected from inaccuracies in the PSF modeling
        """
        x_pos, y_pos = self._pixel_grid.map_coord2pix(ra_pos, dec_pos)
        psf_kernel = self._psf.kernel_point_source
        psf_error_map = self._psf.psf_error_map
        error_map = np.zeros_like(data)
        for i in range(len(x_pos)):
            if fix_psf_error_map is True:
                amp_estimated = amp
            else:
                amp_estimated = kernel_util.estimate_amp(
                    data, x_pos[i], y_pos[i], psf_kernel
                )
            error_map = image_util.add_layer2image(
                error_map, x_pos[i], y_pos[i], psf_error_map * amp_estimated**2
            )
        return error_map



from lenstronomy.ImSim.differential_extinction import DifferentialExtinction
import pytest
import numpy.testing as npt


class TestDifferentialExtinction(object):
    def setup_method(self):
        pass

    def test_extinction(self):
        extinction = DifferentialExtinction(
            optical_depth_model=["GAUSSIAN"], tau0_index=0
        )
        kwargs_extinction = [{"amp": 1, "sigma": 1, "center_x": 0, "center_y": 0}]
        kwargs_special = {"tau0_list": [2, 0]}
        ext = extinction.extinction(
            x=1, y=1, kwargs_special=kwargs_special, kwargs_extinction=kwargs_extinction
        )
        npt.assert_almost_equal(ext, 0.8894965388088921, decimal=8)

        ext = extinction.extinction(
            x=1, y=1, kwargs_special=kwargs_special, kwargs_extinction=None
        )
        npt.assert_almost_equal(ext, 1, decimal=8)

        ext = extinction.extinction(
            x=1, y=1, kwargs_special={}, kwargs_extinction=kwargs_extinction
        )
        npt.assert_almost_equal(ext, 0.9431312415612645, decimal=8)


if __name__ == "__main__":
    pytest.main()



import copy
import numpy as np

import lenstronomy.Util.class_creator as class_creator
from lenstronomy.Plots.model_band_plot import ModelBandPlot
from lenstronomy.Analysis.image_reconstruction import check_solver_error

__all__ = ["ModelPlot"]


class ModelPlot(object):
    """Class that manages the summary plots of a lens model The class uses the same
    conventions as being used in the FittingSequence and interfaces with the ImSim
    module.

    The linear inversion is re-done given the likelihood settings in the init of this
    class (make sure this is the same as you perform the FittingSequence) to make sure
    the linear amplitude parameters are computed as they are not part of the output of
    the FittingSequence results.
    """

    def __init__(
        self,
        multi_band_list,
        kwargs_model,
        kwargs_params,
        image_likelihood_mask_list=None,
        bands_compute=None,
        multi_band_type="multi-linear",
        source_marg=False,
        linear_prior=None,
        arrow_size=0.02,
        cmap_string="gist_heat",
        fast_caustic=True,
        linear_solver=True,
    ):
        """

        :param multi_band_list: list of [[kwargs_data, kwargs_psf, kwargs_numerics], [], ..]
        :param multi_band_type: string, option when having multiple imaging data sets modelled simultaneously.
         Options are:
         - 'multi-linear': linear amplitudes are inferred on single data set
         - 'linear-joint': linear amplitudes ae jointly inferred
         - 'single-band': single band
        :param kwargs_model: model keyword arguments
        :param bands_compute: (optional), bool list to indicate which band to be included in the modeling
        :param image_likelihood_mask_list: list of image likelihood mask
         (same size as image_data with 1 indicating being evaluated and 0 being left out)
        :param kwargs_params: keyword arguments of 'kwargs_lens', 'kwargs_source' etc. as coming as kwargs_result from
         FittingSequence class
        :param source_marg:
        :param linear_prior:
        :param arrow_size:
        :param cmap_string:
        :param fast_caustic: boolean; if True, uses fast (but less accurate) caustic calculation method
        :param linear_solver: bool, if True (default) fixes the linear amplitude parameters 'amp' (avoid sampling) such
         that they get overwritten by the linear solver solution.
        """

        if bands_compute is None:
            bands_compute = [True] * len(multi_band_list)
        if multi_band_type == "single-band":
            multi_band_type = "multi-linear"  # this makes sure that the linear inversion outputs are coming in a list
        self._imageModel = class_creator.create_im_sim(
            multi_band_list,
            multi_band_type,
            kwargs_model,
            bands_compute=bands_compute,
            linear_solver=linear_solver,
            image_likelihood_mask_list=image_likelihood_mask_list,
        )

        kwargs_params_copy = copy.deepcopy(kwargs_params)
        model, error_map, cov_param, param = self._imageModel.image_linear_solve(
            inv_bool=True, **kwargs_params_copy
        )

        if linear_solver is False:
            if len(multi_band_list) > 1:
                raise ValueError(
                    "plotting the solution without the linear solver currently only works with one band."
                )

            im_sim = class_creator.create_im_sim(
                multi_band_list,
                "single-band",
                kwargs_model,
                bands_compute=bands_compute,
                linear_solver=linear_solver,
                image_likelihood_mask_list=image_likelihood_mask_list,
            )
            # overwrite model with initial input without linear solver applied
            model[0] = im_sim.image(**kwargs_params)
            # retrieve amplitude parameters directly from kwargs_list
            param[0] = im_sim.linear_param_from_kwargs(
                kwargs_params["kwargs_source"],
                kwargs_params["kwargs_lens_light"],
                kwargs_params["kwargs_ps"],
            )
        else:
            # overwrite the keyword list with the linear solved 'amp' values
            for key in kwargs_params.keys():
                kwargs_params[key] = kwargs_params_copy[key]

        check_solver_error(param)
        log_l = self._imageModel.likelihood_data_given_model(
            source_marg=source_marg, linear_prior=linear_prior, **kwargs_params
        )

        n_data = self._imageModel.num_data_evaluate
        if n_data > 0:
            print(
                log_l * 2 / n_data,
                "reduced X^2 of all evaluated imaging data combined "
                "(without degrees of freedom subtracted).",
            )

        self._band_plot_list = []
        self._index_list = []
        index = 0
        for i in range(len(multi_band_list)):
            if bands_compute[i] is True:
                if multi_band_type == "joint-linear":
                    param_i = param
                    cov_param_i = cov_param
                else:
                    param_i = param[index]
                    cov_param_i = cov_param[index]

                bandplot = ModelBandPlot(
                    multi_band_list,
                    kwargs_model,
                    model[index],
                    error_map[index],
                    cov_param_i,
                    param_i,
                    copy.deepcopy(kwargs_params),
                    likelihood_mask_list=image_likelihood_mask_list,
                    band_index=i,
                    arrow_size=arrow_size,
                    cmap_string=cmap_string,
                    fast_caustic=fast_caustic,
                )

                self._band_plot_list.append(bandplot)
                self._index_list.append(index)
            else:
                self._index_list.append(-1)
            index += 1

    def _select_band(self, band_index):
        """

        :param band_index: index of imaging band to be plotted
        :return: bandplot() instance of selected band, raises when band is not computed
        """
        i = self._index_list[band_index]
        if i == -1:
            raise ValueError("band %s is not computed or out of range." % band_index)
        i = int(i)
        return self._band_plot_list[i]

    def reconstruction_all_bands(self, **kwargs):
        """

        :param kwargs: arguments of plotting
        :return: 3 x n_data plot with data, model, reduced residual plots of all the images/bands that are being modeled

        """
        n_bands = len(self._band_plot_list)
        import matplotlib.pyplot as plt

        f, axes = plt.subplots(n_bands, 3, figsize=(12, 4 * n_bands))
        if n_bands == 1:  # make sure axis can be called as 2d array
            _axes = np.empty((1, 3), dtype=object)
            _axes[:] = axes
            axes = _axes
        i = 0
        for band_index in self._index_list:
            if band_index >= 0:
                axes[i, 0].set_title("image " + str(band_index))
                self.data_plot(ax=axes[i, 0], band_index=band_index, **kwargs)
                self.model_plot(
                    ax=axes[i, 1], image_names=True, band_index=band_index, **kwargs
                )
                self.normalized_residual_plot(
                    ax=axes[i, 2], v_min=-6, v_max=6, band_index=band_index, **kwargs
                )
                i += 1
        return f, axes

    def data_plot(self, band_index=0, **kwargs):
        """Illustrates data.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.data_plot(**kwargs)

    def model_plot(self, band_index=0, **kwargs):
        """Illustrates model.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.model_plot(**kwargs)

    def convergence_plot(self, band_index=0, **kwargs):
        """Illustrates lensing convergence in data frame.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.convergence_plot(**kwargs)

    def substructure_plot(self, band_index=0, **kwargs):
        """Illustrates substructure in the lens system.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.substructure_plot(**kwargs)

    def normalized_residual_plot(self, band_index=0, **kwargs):
        """Illustrates normalized residuals between data and model fit.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.normalized_residual_plot(**kwargs)

    def absolute_residual_plot(self, band_index=0, **kwargs):
        """Illustrates absolute residuals between data and model fit.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.absolute_residual_plot(**kwargs)

    def source_plot(self, band_index=0, **kwargs):
        """Illustrates reconstructed source (de-lensed de-convolved)

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.source_plot(**kwargs)

    def error_map_source_plot(self, band_index=0, **kwargs):
        """Illustrates surface brightness variance in the reconstruction in the source
        plane.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.error_map_source_plot(**kwargs)

    def magnification_plot(self, band_index=0, **kwargs):
        """Illustrates lensing magnification in the field of view of the data frame.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.magnification_plot(**kwargs)

    def deflection_plot(self, band_index=0, **kwargs):
        """Illustrates lensing deflections on the field of view of the data frame.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.deflection_plot(**kwargs)

    def decomposition_plot(self, band_index=0, **kwargs):
        """Illustrates decomposition of model components.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.decomposition_plot(**kwargs)

    def subtract_from_data_plot(self, band_index=0, **kwargs):
        """Subtracts individual model components from the data.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.subtract_from_data_plot(**kwargs)

    def plot_main(self, band_index=0, **kwargs):
        """Plot a set of 'main' modelling diagnostics.

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.plot_main(**kwargs)

    def plot_separate(self, band_index=0):
        """Plot a set of 'main' modelling diagnostics.

        :param band_index: index of band
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.plot_separate()

    def plot_subtract_from_data_all(self, band_index=0):
        """Plot a set of 'main' modelling diagnostics.

        :param band_index: index of band
        :return: plot instance
        """
        plot_band = self._select_band(band_index)
        return plot_band.plot_subtract_from_data_all()

    def plot_extinction_map(self, band_index=0, **kwargs):
        """

        :param band_index: index of band
        :param kwargs: arguments of plotting
        :return: plot instance of differential extinction map
        """
        plot_band = self._select_band(band_index)
        return plot_band.plot_extinction_map(**kwargs)

    def source(self, band_index=0, **kwargs):
        """

        :param band_index: index of band
        :param kwargs: keyword arguments accessible in model_band_plot.source()
        :return: 2d array of source surface brightness
        """
        plot_band = self._select_band(band_index)
        return plot_band.source(**kwargs)



__author__ = "dgilman"

from lenstronomy.LensModel.lens_model import LensModel
import numpy as np


class MultiplaneFast(object):
    """This class accelerates ray tracing computations in multi plane lensing for
    quadruple image lenses by only computing the deflection from objects in front of the
    main deflector at z_lens one time.

    The first ray tracing computation through the foreground is saved and re-used, but
    it will always have the same shape as the initial x_image, y_image arrays.
    """

    def __init__(
        self,
        x_image,
        y_image,
        z_lens,
        z_source,
        lens_model_list,
        redshift_list,
        astropy_instance,
        param_class,
        foreground_rays,
        tol_source=1e-5,
        numerical_alpha_class=None,
    ):
        """

        :param x_image: x_image to fit
        :param y_image: y_image to fit
        :param z_lens: lens redshift
        :param z_source: source redshift
        :param lens_model_list: list of lens models
        :param redshift_list: list of lens redshifts
        :param astropy_instance: instance of astropy to pass to lens model
        :param param_class: an instance of ParamClass (see documentation in QuadOptimmizer.param_manager)
        :param foreground_rays: (optional) pre-computed foreground rays from a previous iteration, if they are not specified
         they will be re-computed
        :param tol_source: source plane chi^2 sigma
        :param numerical_alpha_class: class for computing numerically tabulated deflection angles
        """

        self.lensModel = LensModel(
            lens_model_list,
            z_lens,
            z_source,
            redshift_list,
            astropy_instance,
            multi_plane=True,
            numerical_alpha_class=numerical_alpha_class,
        )

        lensmodel_list_to_vary = lens_model_list[0 : param_class.to_vary_index]
        redshift_list_to_vary = redshift_list[0 : param_class.to_vary_index]
        lensmodel_list_fixed = lens_model_list[param_class.to_vary_index :]
        redshift_list_fixed = redshift_list[param_class.to_vary_index :]

        self.lens_model_to_vary = LensModel(
            lensmodel_list_to_vary,
            z_lens,
            z_source,
            redshift_list_to_vary,
            cosmo=astropy_instance,
            multi_plane=True,
            numerical_alpha_class=numerical_alpha_class,
        )

        self.lens_model_fixed = LensModel(
            lensmodel_list_fixed,
            z_lens,
            z_source,
            redshift_list_fixed,
            cosmo=astropy_instance,
            multi_plane=True,
            numerical_alpha_class=numerical_alpha_class,
        )

        self._z_lens = z_lens

        self._z_source = z_source

        self._x_image = x_image
        self._y_image = y_image
        self._param_class = param_class

        self._tol_source = tol_source

        self._foreground_rays = foreground_rays

    def chi_square(self, args_lens, *args, **kwargs):
        """

        :param args_lens: array of lens model parameters being optimized, computed from kwargs_lens in a specified
         param_class, see documentation in QuadOptimizer.param_manager
        :return: total chi^2 penalty (source chi^2 + param chi^2), where param chi^2 is computed by the specified
         param_class
        """
        source_plane_penlty = self.source_plane_chi_square(args_lens)

        param_penalty = self._param_class.param_chi_square_penalty(args_lens)

        return source_plane_penlty + param_penalty

    def logL(self, args_lens, *args, **kwargs):
        """

        :param args_lens: array of lens model parameters being optimized, computed from kwargs_lens in a specified
         param_class, see documentation in QuadOptimizer.param_manager
        :return: the log likelihood corresponding to the given chi^2
        """
        chi_square = self.chi_square(args_lens)

        return -0.5 * chi_square

    def source_plane_chi_square(self, args_lens, *args, **kwargs):
        """

        :param args_lens: array of lens model parameters being optimized, computed from kwargs_lens in a specified
         param_class, see documentation in QuadOptimizer.param_manager
        :return: chi2 penalty for the source position (all images must map to the same source coordinate)
        """

        betax, betay = self.ray_shooting_fast(args_lens)

        dx_source = (
            (betax[0] - betax[1]) ** 2
            + (betax[0] - betax[2]) ** 2
            + (betax[0] - betax[3]) ** 2
            + (betax[1] - betax[2]) ** 2
            + (betax[1] - betax[3]) ** 2
            + (betax[2] - betax[3]) ** 2
        )
        dy_source = (
            (betay[0] - betay[1]) ** 2
            + (betay[0] - betay[2]) ** 2
            + (betay[0] - betay[3]) ** 2
            + (betay[1] - betay[2]) ** 2
            + (betay[1] - betay[3]) ** 2
            + (betay[2] - betay[3]) ** 2
        )

        chi_square = 0.5 * (dx_source + dy_source) / self._tol_source**2

        return chi_square

    def ray_shooting_fast(self, args_lens):
        """Performs a ray tracing computation through observed coordinates on the sky
        (self._x_image, self._y_image) to the source plane, returning the final
        coordinates of each ray on the source plane.

        :param args_lens: An array of parameters being optimized. The array is computed
            from a set of key word arguments by an instance of ParamClass (see
            documentation in QuadOptimizer.param_manager)
        :return: the xy coordinate of each ray traced back to the source plane
        """

        # these do not depend on kwargs_lens_array
        x, y, alpha_x, alpha_y = self._ray_shooting_fast_foreground()

        # convert array into new kwargs dictionary
        kw = self._param_class.args_to_kwargs(args_lens)
        index = self._param_class.to_vary_index
        kwargs_lens = kw[0:index]
        # evaluate main deflector deflection angles
        (
            x,
            y,
            alpha_x,
            alpha_y,
        ) = self.lens_model_to_vary.lens_model.ray_shooting_partial(
            x,
            y,
            alpha_x,
            alpha_y,
            self._z_lens,
            self._z_lens,
            kwargs_lens,
            include_z_start=True,
        )

        # ray trace through background halos
        kwargs_lens = kw[index:]
        x, y, _, _ = self.lens_model_fixed.lens_model.ray_shooting_partial(
            x,
            y,
            alpha_x,
            alpha_y,
            self._z_lens,
            self._z_source,
            kwargs_lens,
            check_convention=False,
        )

        beta_x, beta_y = self.lens_model_fixed.lens_model.co_moving2angle_source(x, y)

        return beta_x, beta_y

    def _ray_shooting_fast_foreground(self):
        """Does the ray tracing through the foreground halos only once."""

        if self._foreground_rays is None:
            # These do not depend on the kwargs being optimized for
            kw = self._param_class.kwargs_lens
            index = self._param_class.to_vary_index
            kwargs_lens = kw[index:]

            x0, y0 = np.zeros_like(self._x_image), np.zeros_like(self._y_image)
            (
                x,
                y,
                alpha_x,
                alpha_y,
            ) = self.lens_model_fixed.lens_model.ray_shooting_partial(
                x0,
                y0,
                self._x_image,
                self._y_image,
                z_start=0.0,
                z_stop=self._z_lens,
                kwargs_lens=kwargs_lens,
            )

            self._foreground_rays = (x, y, alpha_x, alpha_y)

        return (
            self._foreground_rays[0],
            self._foreground_rays[1],
            self._foreground_rays[2],
            self._foreground_rays[3],
        )
